
from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.submission_SYNC import Submission

from functools import cached_property

from ...pagination.paginator_chaining_iterator import ImpartedPaginatorChainingIterator
from ...pagination.paginators.wiki_sync1 import WikiPageRevisionsPaginator
from ...pagination.paginators.listing.submission_listing_paginator import SubmissionListingPaginator
from ...models.wiki_SYNC import WikiPage, WikiPageRevision, WikiPageSettings
from ...model_loaders.wiki_SYNC import load_wiki_page, load_wiki_page_settings

class WikiProcedures:
    def __init__(self, client: Client):
        self._client = client

    class _page:
        def __init__(self, outer: WikiProcedures):
            self._outer = outer
            self._client = outer._client

        def read(self, sr: str, page: str) -> WikiPage:
            root = self._client.request('GET', f'/r/{sr}/wiki/{page}')
            return load_wiki_page(root['data'], self._client)

        def edit(self, sr: str, page: str, body: str, *, message: str = '') -> None:
            self._client.request('POST', f'/r/{sr}/api/wiki/edit',
                    data={'page': page, 'content': body, 'reason': message})

        def revert(self, sr: str, page: str, revision: str) -> None:
            self._client.request('POST', f'/r/{sr}/api/wiki/revert',
                    data={'page': page, 'revision': revision})

        def revisions(self, sr: str, page: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingIterator[WikiPageRevisionsPaginator, WikiPageRevision]:
            p = WikiPageRevisionsPaginator(self._client, f'/r/{sr}/wiki/revisions/{page}')
            return ImpartedPaginatorChainingIterator(p, amount)

        def discussions(self, sr: str, page: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingIterator[SubmissionListingPaginator, Submission]:
            p = SubmissionListingPaginator(self._client, f'/r/{sr}/wiki/discussions/{page}')
            return ImpartedPaginatorChainingIterator(p, amount)

        def get_settings(self, sr: str, page: str) -> WikiPageSettings:
            root = self._client.request('GET', f'/r/{sr}/wiki/settings/{page}')
            return load_wiki_page_settings(root['data'], self._client)

        def set_settings(self, sr: str, page: str, *, permlevel: int, unlisted: bool) -> WikiPageSettings:
            root = self._client.request('POST', f'/r/{sr}/wiki/settings/{page}',
                    data={'permlevel': str(permlevel), 'listed': '10'[unlisted]})
            return load_wiki_page_settings(root['data'], self._client)

        def add_editor(self, sr: str, page: str, username: str) -> None:
            self._client.request('POST', f'/r/{sr}/api/wiki/alloweditor/add', data={'username': username})

        def remove_editor(self, sr: str, page: str, username: str) -> None:
            self._client.request('POST', f'/r/{sr}/api/wiki/alloweditor/del', data={'username': username})

    page: cached_property[_page] = cached_property(_page)

    def revisions(self, sr: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingIterator[WikiPageRevisionsPaginator, WikiPageRevision]:
        p = WikiPageRevisionsPaginator(self._client, f'/r/{sr}/wiki/revisions')
        return ImpartedPaginatorChainingIterator(p, amount)

    def list_pages(self, sr: str) -> Sequence[str]:
        root = self._client.request('GET', f'/r/{sr}/wiki/pages')
        return root['data']

    def toggle_revision_hidden(self, sr: str, page: str, revision: str) -> bool:
        root = self._client.request('POST', f'/r/{sr}/api/wiki/hide',
                data={'page': page, 'revision': revision})
        return root['status']




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Iterable
if TYPE_CHECKING:
    from ...request import Request
    from ...response import Response

import sys
import urllib.request
import urllib.parse
import urllib.error
import socket
import json

from ..SYNC import register
from ...session_base_SYNC import SessionBase
from ... import exceptions
from ... import payload
from ...response import UResponse
from ...util.merge_query_params import merge_query_params
from ...util.case_insensitive_dict import CaseInsensitiveDict

def _generate_request_kwargs(r: Request) -> Iterable[tuple[str, Any]]:
    yield ('method', r.verb)
    url = merge_query_params(r.uri, r.params)
    yield ('url', url)

    headers = dict(r.headers)
    pld = r.payload
    if pld is None:
        pass

    elif isinstance(pld, payload.Bytes):
        pld.apply_content_type(headers)
        yield ('data', pld.data)

    elif isinstance(pld, payload.Text):
        pld.apply_content_type(headers)
        yield ('data', pld.text.encode())

    elif isinstance(pld, payload.JSON):
        yield ('data', json.dumps(pld.json).encode())
        headers['Content-Type'] = 'application/json'

    elif isinstance(pld, payload.URLEncodedFormData):
        yield ('data', urllib.parse.urlencode(pld.data).encode())

    elif isinstance(pld, payload.MultipartFormData):
        raise Exception('multipart payload not supported by python urllib')

    else:
        raise Exception('unsupported payload type')

    yield ('headers', headers)


class Session(SessionBase):
    def __init__(self) -> None:
        super().__init__()

    def send(self, request: Request, *, timeout: float = -2) -> Response:
        etv = self._get_effective_timeout_value(timeout)
        kwargs = dict(_generate_request_kwargs(request))
        req = urllib.request.Request(**kwargs)
        t = None if etv == -1 else etv
        try:
            with urllib.request.urlopen(req, timeout=t) as response:
                content = response.read()
        except socket.timeout as cause:
            raise exceptions.TimeoutException from cause
        except urllib.error.URLError as cause:
            if str(cause.reason) == 'timed out':
                raise exceptions.TimeoutException from cause
            raise
        except Exception as cause:
            raise exceptions.TransportError from cause

        return UResponse(
            status=response.status,
            headers=CaseInsensitiveDict(dict(response.headers)),
            data=content,
            underlying_object=response,
        )

    def close(self) -> None:
        pass

def new_session() -> Session:
    return Session()

name: str = 'python-urllib'
version: str = '%d.%d' % sys.version_info[:2]
register(
    adaptor_module_name=__name__,
    name=name,
    version=version,
    new_session=new_session,
)




from __future__ import annotations
from typing import TypeVar, Sequence, Generic, AsyncIterator, Optional

T = TypeVar('T')

class AsyncPaginator(Generic[T]):
    def __init__(self, *, limit: Optional[int] = None) -> None:
        self.limit: Optional[int] = limit

    async def __aiter__(self) -> AsyncIterator[Sequence[T]]:
        while x := await self.fetch():
            yield x

    async def fetch(self) -> Sequence[T]:
        raise NotImplementedError

class OffsetAsyncPaginator(AsyncPaginator[T]):
    def __init__(self, *, limit: Optional[int] = None, offset: Optional[int] = None) -> None:
        super().__init__(limit=limit)
        self.offset: Optional[int] = offset

class CursorAsyncPaginator(AsyncPaginator[T]):
    def get_cursor(self) -> str:
        raise NotImplementedError

    def set_cursor(self, value: str) -> None:
        raise NotImplementedError


class MoreAvailableAsyncPaginator(AsyncPaginator[T]):
    async def __aiter__(self) -> AsyncIterator[Sequence[T]]:
        if x := await self.fetch():
            yield x
        while self.more_available() and (x := await self.fetch()):
            yield x

    def more_available(self) -> bool:
        raise NotImplementedError

    def set_more_available_flag(self, value: bool) -> None:
        raise NotImplementedError


class Bidirectional:
    direction: bool

class Resettable:
    def reset(self) -> None:
        raise NotImplementedError




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .typedefs import ClientCredentials
    from ..http.requestor_ASYNC import Requestor

from ..http.request import make_request
from .utils import apply_basic_auth

class TokenRevocationClient:
    def __init__(self, requestor: Requestor, uri: str,
            client_credentials: ClientCredentials):
        self.requestor: Requestor = requestor
        self.uri: str = uri
        self.client_credentials: tuple[str, str] = client_credentials

    async def revoke_token(self, token: str, token_type_hint: str = '') -> None:
        data = {'token': token}
        if token_type_hint:
            data['token_type_hint'] = token_type_hint

        r = make_request('POST', self.uri, data=data)
        apply_basic_auth(r, *self.client_credentials)
        resp = await self.requestor.send(r)
        resp.raise_for_status()

    async def revoke_access_token(self, token: str) -> None:
        await self.revoke_token(token, 'access_token')

    async def revoke_refresh_token(self, token: str) -> None:
        await self.revoke_token(token, 'refresh_token')




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Iterable, TypeVar, Sequence
if TYPE_CHECKING:
    from ..client_SYNC import Client

from .pushshift_paginator import PushshiftPaginator, PushshiftDocumentPaginator, PushshiftDocumentIDPaginator
from ..models.pushshift_document import PushshiftDocument

T = TypeVar('T')

class _CommentSearchPaginatorMixin(PushshiftPaginator[T]):
    _query: str
    _author: str
    _subreddit: str

    def generate_doseq_params(self) -> Iterable[tuple[str, Sequence[str]]]:
        yield from super().generate_doseq_params()
        if self._query:
            yield ('q', (self._query,))
        if self._author:
            yield ('author', (self._author,))
        if self._subreddit:
            yield ('subreddit', (self._subreddit,))


class CommentSearchDocumentPaginator(
    _CommentSearchPaginatorMixin[PushshiftDocument],
    PushshiftDocumentPaginator,
):
    def __init__(self,
        *,
        client: Client,
        uri: str,
        limit: Optional[int] = 100,
        time_range: tuple[Optional[int], Optional[int]] = (None, None),
        ascending: bool = False,
        fields: Iterable[str] = (),
        query: str = '',
        author: str = '',
        subreddit: str = '',
    ) -> None:
        super().__init__(
            client=client,
            uri=uri,
            limit=limit,
            time_range=time_range,
            ascending=ascending,
            fields=fields,
        )
        self._query: str = query
        self._author: str = author
        self._subreddit: str = subreddit

class CommentSearchDocumentIDPaginator(
    _CommentSearchPaginatorMixin[int],
    PushshiftDocumentIDPaginator,
):
    def __init__(self,
        *,
        client: Client,
        uri: str,
        limit: Optional[int] = 100,
        time_range: tuple[Optional[int], Optional[int]] = (None, None),
        ascending: bool = False,
        query: str = '',
        author: str = '',
        subreddit: str = '',
    ) -> None:
        super().__init__(
            client=client,
            uri=uri,
            limit=limit,
            time_range=time_range,
            ascending=ascending,
        )
        self._query: str = query
        self._author: str = author
        self._subreddit: str = subreddit




from __future__ import annotations
from typing import Mapping, Any, Sequence

from datetime import datetime, timezone
from functools import cached_property

from ..auth.const import AUTHORIZATION_BASE_URL

class BaseCustomFeed:
    def __init__(self, d: Mapping[str, Any]):
        self.d: Mapping[str, Any] = d
        self.title: str = d['display_name']
        self.name: str = d['name']
        self.description: str = d['description_md']
        self.description_html: str = d['description_html']
        self.subscriber_count: int = d['num_subscribers']
        self.icon_url: str = d['icon_url']
        self.subreddit_names: Sequence[str] = [o['name'] for o in d['subreddits']]
        self.created_ut: int = int(d['created_utc'])
        self.created_at: datetime = datetime.fromtimestamp(self.created_ut, timezone.utc)
        self.private: bool = d['visibility'] == 'private'
        self.nsfw: bool = d['over_18']
        self.rel_permalink: str = d['path']
        self.permalink: str = AUTHORIZATION_BASE_URL + d['path']
        self.copied_from: str = d['copied_from']
        self.owner: str = d['owner']

        full_id36: str = d['owner_id']
        _, _, id36 = full_id36.partition('_')
        self.owner_id36: str = id36
        self.owner_id: int = int(id36, 36)

    class _me:
        def __init__(self, outer: BaseCustomFeed):
            d = outer.d
            self.can_edit: bool = d['can_edit']
            self.favorited: bool = d['is_favorited']
            self.is_subscribed: bool = d['is_subscriber']

    me: cached_property[_me] = cached_property(_me)




from __future__ import annotations
from typing import Mapping, Any

class UserSummary:
    def __init__(self, d: Mapping[str, Any], id36: str):
        self.d: Mapping[str, Any] = d
        self.id36: str = id36
        self.id: int = int(id36, 36)
        self.name: str = d['name']
        self.created_ut: int = int(d['created_utc'])
        self.submission_karma: int = d['link_karma']
        self.comment_karma: int = d['comment_karma']
        self.profile_img: str = d['profile_img']
        self.profile_color: str = d['profile_color']
        self.nsfw: bool = d['profile_over_18']



#!/usr/bin/env python3
"""
Run refresh_token.py with the `--access-token-only` flag.
"""
import runpy
runpy.run_module('redditwarp.cli.refresh_token', init_globals={
        '?access_token_only': True, '?description': __doc__})




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Iterable, Sequence, Any, Mapping
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.subreddit_ASYNC import Subreddit
    from ...models.comment_ASYNC import ExtraSubmissionFieldsComment
    from ...models.subreddit_rules import SubredditRules
    from ...models.moderator_list import ModeratorListItem

from ...model_loaders.subreddit_ASYNC import load_subreddit, load_potentially_inaccessible_subreddit
from ...model_loaders.subreddit_rules import load_subreddit_rules
from ...model_loaders.moderator_list import load_moderator_list_item
from ...util.base_conversion import to_base36
from ...iterators.chunking import chunked
from ...iterators.call_chunk_chaining_async_iterator import CallChunkChainingAsyncIterator
from ...iterators.async_call_chunk import AsyncCallChunk
from ...pagination.paginator_chaining_async_iterator import ImpartedPaginatorChainingAsyncIterator
from ...pagination.paginators.listing.comment_listing_async_paginator import ExtraSubmissionFieldsCommentListingAsyncPaginator
from ...pagination.paginators.subreddit_async1 import SearchSubredditsAsyncPaginator
from ... import exceptions
from ... import http
from ...http.util.json_load import json_loads_response
from .fetch_ASYNC import Fetch
from .get_ASYNC import Get
from .pull_ASYNC import Pull
from .pulls_ASYNC import Pulls

class SubredditProcedures:
    def __init__(self, client: Client):
        self._client = client
        self.get: Get = Get(client)
        self.fetch: Fetch = Fetch(self, client)
        self.pull: Pull = Pull(client)
        self.pulls: Pulls = Pulls(client)

    async def get_by_name(self, name: str) -> Optional[Subreddit]:
        try:
            root = await self._client.request('GET', f'/r/{name}/about')
        except http.exceptions.StatusCodeException as e:
            if e.status_code == 404:
                return None
            raise
        if root['kind'] == 'Listing':
            return None
        return load_subreddit(root['data'], self._client)

    async def fetch_by_name(self, name: str) -> Subreddit:
        root = await self._client.request('GET', f'/r/{name}/about')
        if root['kind'] == 'Listing':
            raise exceptions.NoResultException('target not found')
        return load_subreddit(root['data'], self._client)

    async def get_potentially_inaccessible_subreddit_by_name(self, name: str) -> Optional[object]:
        root = await self._client.request('GET', '/api/info', params={'sr_name': name})
        if children := root['data']['children']:
            return load_potentially_inaccessible_subreddit(children[0]['data'], self._client)
        return None

    async def fetch_potentially_inaccessible_subreddit_by_name(self, name: str) -> object:
        v = await self.get_potentially_inaccessible_subreddit_by_name(name)
        if v is None:
            raise exceptions.NoResultException('target not found')
        return v

    def bulk_fetch_potentially_inaccessible_subreddits(self, ids: Iterable[int]) -> CallChunkChainingAsyncIterator[object]:
        async def mass_fetch(ids: Sequence[int]) -> Sequence[object]:
            id36s = map(to_base36, ids)
            full_id36s = map('t5_'.__add__, id36s)
            ids_str = ','.join(full_id36s)
            root = await self._client.request('GET', '/api/info', params={'id': ids_str})
            return [load_potentially_inaccessible_subreddit(i['data'], self._client) for i in root['data']['children']]

        return CallChunkChainingAsyncIterator(AsyncCallChunk(mass_fetch, chunk) for chunk in chunked(ids, 100))

    def bulk_fetch_potentially_inaccessible_subreddits_by_name(self, names: Iterable[str]) -> CallChunkChainingAsyncIterator[object]:
        async def mass_fetch(names: Sequence[str]) -> Sequence[object]:
            root = await self._client.request('GET', '/api/info', params={'sr_name': ','.join(names)})
            return [load_potentially_inaccessible_subreddit(i['data'], self._client) for i in root['data']['children']]

        return CallChunkChainingAsyncIterator(AsyncCallChunk(mass_fetch, chunk) for chunk in chunked(names, 100))

    def pull_new_comments(self, sr: str, amount: Optional[int] = None) -> ImpartedPaginatorChainingAsyncIterator[ExtraSubmissionFieldsCommentListingAsyncPaginator, ExtraSubmissionFieldsComment]:
        p = ExtraSubmissionFieldsCommentListingAsyncPaginator(self._client, f'/r/{sr}/comments')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    async def get_settings(self, sr: str) -> Mapping[str, Any]:
        root = await self._client.request('GET', f'/r/{sr}/about/edit')
        if root['kind'] != 'subreddit_settings':
            raise exceptions.NoResultException('target not found')
        return root['data']

    async def update_settings(self, data: Mapping[str, Any]) -> None:
        await self._client.request('PATCH', '/api/v1/subreddit/update_settings', json=data)

    async def DEFUNCT__trending_subreddit_names(self) -> Sequence[str]:
        req = self._client.http.session.make_request('GET', 'https://reddit.com/api/trending_subreddits.json')
        resp = await self._client.http.send(req)
        jdata = json_loads_response(resp)
        return jdata['subreddit_names']

    async def subscribe_by_name(self, sr: str) -> None:
        await self._client.request('POST', '/api/subscribe', data={'action': 'sub', 'sr_name': sr})

    async def subscribe_by_id(self, idn: int) -> None:
        id36 = to_base36(idn)
        full_id36 = 't5_' + id36
        await self._client.request('POST', '/api/subscribe', data={'action': 'sub', 'sr': full_id36})

    async def unsubscribe_by_name(self, sr: str) -> None:
        await self._client.request('POST', '/api/subscribe', data={'action': 'unsub', 'sr_name': sr})

    async def unsubscribe_by_id(self, idn: int) -> None:
        id36 = to_base36(idn)
        full_id36 = 't5_' + id36
        await self._client.request('POST', '/api/subscribe', data={'action': 'unsub', 'sr': full_id36})

    async def get_rules(self, sr: str) -> SubredditRules:
        root = await self._client.request('GET', f'/r/{sr}/about/rules')
        if 'rules' not in root:
            raise exceptions.NoResultException('target not found')
        return load_subreddit_rules(root)

    async def get_post_requirements(self, sr: str) -> Mapping[str, Any]:
        return await self._client.request('GET', f'/api/v1/{sr}/post_requirements')

    def search(self, query: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingAsyncIterator[SearchSubredditsAsyncPaginator, Subreddit]:
        if not query:
            raise ValueError('query cannot be empty')
        p = SearchSubredditsAsyncPaginator(self._client, '/subreddits/search', query)
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    async def search_names(self, name: str) -> Sequence[str]:
        root = await self._client.request('GET', '/api/search_reddit_names', params={'query': name})
        return root['names']

    async def exists(self, name: str) -> bool:
        try:
            await self._client.request('GET', '/api/search_reddit_names', params={'query': name, 'exact': '1'})
        except http.exceptions.StatusCodeException as e:
            if e.status_code == 404:
                return False
            raise
        return True

    async def list_moderators(self, sr: str) -> Sequence[ModeratorListItem]:
        root = await self._client.request('GET', f'/r/{sr}/about/moderators')
        return [load_moderator_list_item(d) for d in root['data']['children']]




from __future__ import annotations
from typing import Optional, Mapping

import urllib.parse

_NO_VALUE_PARAM_PLACEHOLDER = 'NO_VALUE_PARAM_PLACEHOLDER__d07d4b7d-d881-4217-9bca-42ed0c6a3d04'

def merge_query_params(uri: str, params: Mapping[str, Optional[str]]) -> str:
    urlparts = list(urllib.parse.urlparse(uri))
    query_dict = urllib.parse.parse_qs(urlparts[4])

    none_keys = []
    for k, v in params.items():
        if v is None:
            none_keys.append(k)
            v = _NO_VALUE_PARAM_PLACEHOLDER
        query_dict.setdefault(k, []).append(v)

    query = urllib.parse.urlencode(query_dict, doseq=True)

    for k in none_keys:
        query = query.replace(f'{k}={_NO_VALUE_PARAM_PLACEHOLDER}', k)

    urlparts[4] = query
    return urllib.parse.urlunparse(urlparts)




from __future__ import annotations
from typing import TypeVar, Generic, Optional, Sequence

from dataclasses import dataclass

T = TypeVar('T')
TChild = TypeVar('TChild')

@dataclass(repr=False, eq=False)
class Node(Generic[T]):
    value: T

@dataclass(repr=False, eq=False)
class TreeNode(Node[T], Generic[T, TChild]):
    pass

@dataclass(repr=False, eq=False)
class GeneralTreeNode(TreeNode[T, TChild]):
    children: Sequence[TChild]

@dataclass(repr=False, eq=False)
class BinaryTreeNode(TreeNode[T, TChild]):
    left: Optional[TChild]
    right: Optional[TChild]




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence, Iterable, Any, TypeVar
if TYPE_CHECKING:
    from ..client_SYNC import Client

import urllib.parse

from ...pagination.paginator import Paginator
from ..models.pushshift_document import PushshiftDocument

class MissingExpectedFieldException(Exception):
    pass


T = TypeVar('T')

class PushshiftPaginator(Paginator[T]):
    def __init__(self,
        *,
        client: Client,
        uri: str,
        limit: Optional[int] = 100,
        time_range: tuple[Optional[int], Optional[int]] = (None, None),
        ascending: bool = False,
        fields: Iterable[str] = (),
    ) -> None:
        super().__init__(limit=limit)
        self.client: Client = client
        self.uri: str = uri
        self._ascending: bool = ascending
        self._fields: Iterable[str] = fields
        self._time_range_start: Optional[int] = time_range[0]
        self._time_range_stop: Optional[int] = time_range[1]
        self._before_created_utc: Optional[int] = None
        self._after_created_utc: Optional[int] = None
        self._previous_page_item_ids: Sequence[int] = ()

    def generate_doseq_params(self) -> Iterable[tuple[str, Sequence[str]]]:
        if self.limit is not None:
            yield ('limit', (str(self.limit),))

        after: Optional[int] = None if (x := self._time_range_start) is None else x - 1
        before: Optional[int] = self._time_range_stop
        if self._ascending:
            yield ('sort', 'asc')

            if (x := self._after_created_utc):
                after = x - 1
        else:
            yield ('sort', 'desc')

            if (x := self._before_created_utc):
                before = x + 1

        if after is not None:
            yield ('after', (str(after),))
        if before is not None:
            yield ('before', (str(before),))

        if self._fields:
            yield ('fields', list(self._fields))

    def _fetch_data(self) -> Any:
        # Pushshift does not support URL encoded commas in query strings.
        # https://github.com/pushshift/api/issues/26
        query_pairs = list(self.generate_doseq_params())
        query_string = urllib.parse.urlencode(query_pairs, doseq=True)
        uri = '%s?%s' % (self.uri, query_string)

        root = self.client.request('GET', uri)
        documents = root['data']

        previous_page_item_ids_set = set(self._previous_page_item_ids)
        documents = [
            d
            for d in documents
            if int(d.get('id'), 36) not in previous_page_item_ids_set
        ]
        self._previous_page_item_ids = [
            int(x, 36)
            for d in documents
            if (x := d.get('id')) is not None
        ]

        if documents:
            first_created_utc = documents[0].get('created_utc')
            last_created_utc = documents[-1].get('created_utc')
            new_before_created_utc: Optional[int] = last_created_utc
            new_after_created_utc: Optional[int] = first_created_utc
            if self._ascending:
                new_after_created_utc, new_before_created_utc = new_before_created_utc, new_after_created_utc

            if (
                self._before_created_utc is not None and new_before_created_utc is None
                or
                self._after_created_utc is not None and new_after_created_utc is None
            ):
                raise MissingExpectedFieldException('a document is missing the `created_utc` field')

            self._before_created_utc = new_before_created_utc
            self._after_created_utc = new_after_created_utc

        return documents


class PushshiftDocumentPaginator(PushshiftPaginator[PushshiftDocument]):
    def fetch(self) -> Sequence[PushshiftDocument]:
        data = self._fetch_data()
        return [PushshiftDocument(d) for d in data]

class PushshiftDocumentIDPaginator(PushshiftPaginator[int]):
    _FIELDS: Sequence[str] = ('id', 'created_utc')

    def __init__(self,
        *,
        client: Client,
        uri: str,
        limit: Optional[int] = 100,
        time_range: tuple[Optional[int], Optional[int]] = (None, None),
        ascending: bool = False,
    ) -> None:
        super().__init__(
            client=client,
            uri=uri,
            limit=limit,
            time_range=time_range,
            ascending=ascending,
            fields=self._FIELDS,
        )

    def fetch(self) -> Sequence[int]:
        data = self._fetch_data()
        return [int(d['id'], 36) for d in data]




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.user_relationship_item import (
        UserRelationshipItem,
        BannedUserRelationshipItem,
    )

from ...pagination.paginator_chaining_async_iterator import ImpartedPaginatorChainingAsyncIterator
from ...pagination.paginators.moderation.async1 import (
    UserRelationshipItemListingAsyncPaginator,
    BannedUserRelationshipItemListingAsyncPaginator,
)

class LegacyPullUsers:
    def __init__(self, client: Client) -> None:
        self._client = client

    def contributors(self, sr: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingAsyncIterator[UserRelationshipItemListingAsyncPaginator, UserRelationshipItem]:
        p = UserRelationshipItemListingAsyncPaginator(self._client, f'/r/{sr}/about/contributors')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def wiki_contributors(self, sr: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingAsyncIterator[UserRelationshipItemListingAsyncPaginator, UserRelationshipItem]:
        p = UserRelationshipItemListingAsyncPaginator(self._client, f'/r/{sr}/about/wikicontributors')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def banned(self, sr: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingAsyncIterator[BannedUserRelationshipItemListingAsyncPaginator, BannedUserRelationshipItem]:
        p = BannedUserRelationshipItemListingAsyncPaginator(self._client, f'/r/{sr}/about/banned')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def muted(self, sr: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingAsyncIterator[UserRelationshipItemListingAsyncPaginator, UserRelationshipItem]:
        p = UserRelationshipItemListingAsyncPaginator(self._client, f'/r/{sr}/about/muted')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def wiki_banned(self, sr: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingAsyncIterator[BannedUserRelationshipItemListingAsyncPaginator, BannedUserRelationshipItem]:
        p = BannedUserRelationshipItemListingAsyncPaginator(self._client, f'/r/{sr}/about/wikibanned')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)




from __future__ import annotations

from enum import IntEnum

class Side(IntEnum):
    NONE = 0
    CLIENT = 1
    SERVER = 2

class ConnectionState(IntEnum):
    CONNECTING = 0
    OPEN = 1
    CLOSING = 2
    CLOSED = 3

class Opcode(IntEnum):
    CONTINUATION = 0x0
    TEXT = 0x1
    BINARY = 0x2
    CLOSE = 0x8
    PING = 0x9
    PONG = 0xA

'''
def is_data_frame_opcode(v: int) -> bool:
    return (v & 0x8) == 0

def is_control_frame_opcode(v: int) -> bool:
    return (v & 0x8) != 0
'''

AUTHORITATIVE_CLOSE_FRAME_CLOSE_CODES: set[int] = {
    1000,
    1001,
    1002,
    1003,
    1007,
    1008,
    1009,
    1010,
    1011,
    1012,
    1013,
    1014,
}

FORBIDDEN_CLOSE_FRAME_CLOSE_CODES: set[int] = {
    1005,
    1006,
    1015,
}




from __future__ import annotations
from typing import Mapping, Any, Optional

from datetime import datetime, timezone

from .artifact import Artifact

class ModeratedSubreddit(Artifact):
    class Me:
        def __init__(self, d: Mapping[str, Any]):
            self.is_subscribed: bool = d['user_is_subscriber']

    def __init__(self, d: Mapping[str, Any]):
        super().__init__(d)
        full_id36: str = d['name']
        self.id36: str = full_id36[3:]
        self.id: int = int(self.id36, 36)
        self.name: str = d['display_name']
        self.openness: str = d['subreddit_type']
        self.subscriber_count: int = d['subscribers']
        self.title: str = d['title']
        self.nsfw: bool = d['over_18']
        self.icon_img: str = d['icon_img']

        self.me: Optional[ModeratedSubreddit.Me] = None
        if 'user_is_subscriber' in d:
            self.me = self.Me(d)

class ModeratedUserSubreddit(ModeratedSubreddit):
    pass

class ModeratedRegularSubreddit(ModeratedSubreddit):
    def __init__(self, d: Mapping[str, Any]):
        super().__init__(d)
        self.created_ut: int = int(d['created_utc'])
        self.created_at: datetime = datetime.fromtimestamp(self.created_ut, timezone.utc)




from __future__ import annotations
from typing import TypeVar, Mapping, MutableMapping, Iterator, IO, cast, Tuple, Optional, Any

from pprint import PrettyPrinter

V = TypeVar('V')

class CaseInsensitiveDict(MutableMapping[str, V]):
    """A case-folding-case-preserving dictionary."""

    def __init__(self, data: Optional[Mapping[str, V]] = None, **kwargs: V) -> None:
        self._store: MutableMapping[str, Tuple[str, V]] = {}
        if data is None:
            data = {}
        self.update(data, **kwargs)

    def __repr__(self) -> str:
        if self._store:
            return '%s(%r)' % (type(self).__name__, dict(self))
        return type(self).__name__ + '()'

    def __iter__(self) -> Iterator[str]:
        return (k for k, _ in self._store.values())

    def __len__(self) -> int:
        return len(self._store)

    def __setitem__(self, key: str, value: V) -> None:
        self._store[key.casefold()] = (key, value)

    def __getitem__(self, key: str) -> V:
        return self._store[key.casefold()][1]

    def __delitem__(self, key: str) -> None:
        del self._store[key.casefold()]

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Mapping):
            if not isinstance(other, self.__class__):
                other = cast(Mapping[str, V], CaseInsensitiveDict(other))
            return other.items() == self.items()
        return NotImplemented

    def __getstate__(self) -> Mapping[str, Tuple[str, V]]:
        return self._store

    def __setstate__(self, state: Mapping[str, Tuple[str, V]]) -> None:
        object.__setattr__(self, '_store', state)

    @staticmethod
    def _pprint(
        printer: PrettyPrinter,
        obj: Mapping[str, V],
        stream: IO[str],
        indent: int,
        allowance: int,
        context: Mapping[int, Any],
        level: int,
    ) -> None:
        cls_name = obj.__class__.__name__
        stream.write(cls_name + '(')
        printer._format(  # type: ignore[attr-defined]
            dict(obj),
            stream,
            indent + len(cls_name) + 1,
            allowance + 1,
            context,
            level,
        )
        stream.write(')')

    if isinstance(getattr(PrettyPrinter, '_dispatch', None), dict):
        PrettyPrinter._dispatch[__repr__] = _pprint.__func__  # type: ignore[attr-defined]




from __future__ import annotations
from typing import TYPE_CHECKING, MutableSequence, cast, Iterator
if TYPE_CHECKING:
    from redditwarp.models.comment_tree_SYNC import ICommentSubtreeTreeNode, CommentSubtreeTreeNode, T

import sys
import shutil
from collections import deque

import redditwarp
from redditwarp.util.extract_id_from_url import extract_submission_id_from_url
from redditwarp.models.comment_SYNC import Comment

ALGORITHM_CHOICES = {
    'iterative_depth_first_search': 'dfs',
    'depth_first_search': 'dfs',
    'dfs': 'dfs',
    'recursive_depth_first_search': 'recursive_depth_first_search',
    'breadth_first_search': 'bfs',
    'bfs': 'bfs',
}
COMMENT_SORT_CHOICES = 'confidence top new controversial old random qa live'.split()

import argparse
class Formatter(argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter): pass
parser = argparse.ArgumentParser(description=__doc__, formatter_class=Formatter)
parser._optionals.title = __import__('gettext').gettext('named arguments')
add = parser.add_argument
add('target', nargs='?')
add('--access-token', metavar='ACCESS_TOKEN', dest='access_token_opt')
add('--target', metavar='TARGET', dest='target_opt', help=argparse.SUPPRESS)
add('--base', metavar='N', dest='base_opt', type=int, default=36)
add('--algo', '--algorithm', choices=ALGORITHM_CHOICES, default='dfs', dest='algo', metavar='')
add('--comment-sort', dest='comment_sort', choices=COMMENT_SORT_CHOICES)
args = parser.parse_args()

access_token: str = args.access_token_opt or ''
target: str = args.target or args.target_opt or input('Target: ')
chosen_algo: str = args.algo
comment_sort: str = args.comment_sort
base: int = args.base_opt


def recursive_depth_first_search(node: ICommentSubtreeTreeNode) -> Iterator[tuple[int, Comment]]:
    def dfs(
        root: CommentSubtreeTreeNode[T],
        level: int = 0,
    ) -> Iterator[tuple[int, Comment]]:
        value = root.value
        if isinstance(value, Comment):
            yield (level, value)

        for child in root.children:
            yield from dfs(child, level + 1)

        if root.more:
            yield from dfs(root.more(), level)

    return dfs(cast("CommentSubtreeTreeNode[object]", node))

def depth_first_search(node: ICommentSubtreeTreeNode) -> Iterator[tuple[int, Comment]]:
    stack: MutableSequence[ICommentSubtreeTreeNode] = deque([node])
    levels = deque([0])
    while stack:
        node = cast("CommentSubtreeTreeNode[object]", stack.pop())
        level = levels.pop()

        value = node.value
        if isinstance(value, Comment):
            yield (level, value)

        if node.more:
            stack.append(node.more())
            levels.append(level)

        stack.extend(reversed(node.children))
        levels.extend([level + 1] * len(node.children))

def breadth_first_search(node: ICommentSubtreeTreeNode) -> Iterator[tuple[int, Comment]]:
    level = 0
    queue: MutableSequence[ICommentSubtreeTreeNode] = deque([node])
    while queue:
        batch = deque(queue)
        queue.clear()
        while batch:
            node = cast("CommentSubtreeTreeNode[object]", batch.popleft())
            if node.value is None:
                if node.more:
                    batch.appendleft(node.more())
                batch.extendleft(reversed(node.children))
                continue

            value = node.value
            if isinstance(value, Comment):
                yield (level, value)

            queue.extend(node.children)
            if node.more:
                queue.append(node.more())

        level += 1


algo = ALGORITHM_CHOICES[chosen_algo]
traversal = {
    'recursive_depth_first_search': recursive_depth_first_search,
    'dfs': depth_first_search,
    'bfs': breadth_first_search,
}[algo]

client = (
    redditwarp.SYNC.Client.from_access_token(access_token)
    if access_token else
    redditwarp.SYNC.Client()
)

client.http.user_agent += " redditwarp.cli.comment_tree"

idn = 0
if target.isalnum():
    try:
        idn = int(target, base)
    except ValueError as e:
        print(f'{e.__class__.__name__}: {e}', file=sys.stderr)
        sys.exit(1)
else:
    try:
        idn = extract_submission_id_from_url(target)
    except ValueError:
        print('Could not extract comment ID from URL.', file=sys.stderr)
        sys.exit(1)

tree_node = client.p.comment_tree.get(idn, sort=comment_sort)
if tree_node is None:
    print('Submission not found', file=sys.stderr)
    sys.exit(1)

m = submission = tree_node.value

print(f'''\
{m.permalink}
{m.score} :: {m.title}
comments/{m.id36} by u/{m.author_name} to r/{m.subreddit.name}
Submitted at {m.created_at.astimezone().ctime()}{' *' if m.edited else ''}
''')

columns, _lines = shutil.get_terminal_size()

for depth, comment in traversal(tree_node):
    c = comment
    body_text = repr(c.body)
    line = f"{depth*'.'} u/{c.author_name} | {body_text}"
    print(line[:columns])




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Mapping
if TYPE_CHECKING:
    from ..client_SYNC import Client

from dataclasses import dataclass

from .wiki_base import (
    BaseWikiPageRevisionAuthorUser,
    GBaseWikiPage,
    GBaseWikiPageRevision,
    GBaseWikiPageSettings,
)

class WikiPageRevisionAuthorUser(BaseWikiPageRevisionAuthorUser):
    def __init__(self, d: Mapping[str, Any], client: Client):
        super().__init__(d)
        self.client: Client = client

@dataclass(repr=False, eq=False)
class WikiPage(GBaseWikiPage[WikiPageRevisionAuthorUser]):
    pass

@dataclass(repr=False, eq=False)
class WikiPageRevision(GBaseWikiPageRevision[WikiPageRevisionAuthorUser]):
    pass

@dataclass(repr=False, eq=False)
class WikiPageSettings(GBaseWikiPageSettings[WikiPageRevisionAuthorUser]):
    pass




from __future__ import annotations
from typing import TypeVar, Iterable, Iterator

E = TypeVar('E')

class UnfalteringChainingIterator(Iterator[E]):
    """Like `itertools.chain.from_iterable()` but is able to continue when
    an exception occurs during iteration.

    Also has a `self.current_iter` attribute.
    """

    def __init__(self, source: Iterable[Iterable[E]]) -> None:
        self._iterator = iter(source)
        self.current_iter: Iterator[E] = iter(())

    def __iter__(self) -> Iterator[E]:
        return self

    def __next__(self) -> E:
        while True:
            for elem in self.current_iter:
                return elem
            self.current_iter = iter(next(self._iterator))




from __future__ import annotations

import uuid


_REDDITWARP_CLIENT_ID: str = 'Jirijsc5rRo79AzJBUttyQ'

def get_redditwarp_client_id() -> str:
    return _REDDITWARP_CLIENT_ID


_device_id: str = ''

def get_device_id() -> str:
    global _device_id
    if not _device_id:
        _device_id = str(uuid.uuid1())
    return _device_id




from __future__ import annotations
from typing import TYPE_CHECKING, TypeVar, Mapping, Optional, Union, Any
if TYPE_CHECKING:
    from collections.abc import MutableSequence
    from types import TracebackType
    from contextlib import AbstractContextManager
    from .request import Request
    from .response import Response
    from .payload import RequestFiles

import contextvars
from collections import deque

from .requestor_SYNC import Requestor
from .request import make_request

T = TypeVar('T')

DEFAULT_TIMEOUT: float = 100.

class SessionBase(Requestor):
    @staticmethod
    def make_request(
        verb: str,
        uri: str,
        *,
        params: Optional[Mapping[str, str]] = None,
        headers: Optional[Mapping[str, str]] = None,
        data: Optional[Union[Mapping[str, str], bytes]] = None,
        json: Any = None,
        files: Optional[RequestFiles] = None,
    ) -> Request:
        return make_request(
            verb,
            uri,
            params=params,
            headers=headers,
            data=data,
            json=json,
            files=files,
        )

    def __init__(self) -> None:
        self.timeout: float = DEFAULT_TIMEOUT
        self._veto_timeout: contextvars.ContextVar[float] = contextvars.ContextVar('veto_timeout', default=-2.)

    def __enter__(self: T) -> T:
        return self

    def __exit__(self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        self.close()
        return None

    def send(self, request: Request, *, timeout: float = -2) -> Response:
        """
        Parameters
        ----------
        timeout: Optional[float]
            The connect timeout. The number of seconds the client will
            wait to establish a connection to the server.

            A None value will use a default that is specific to the transport
            adaptor. A negative number will wait an infinite amount of time.
        aux_info: Optional[Mapping[Any, Any]]
            Additional information to be consumed by a custom :class:`SessionBase` class.
        """
        raise NotImplementedError

    def request(self,
        verb: str,
        uri: str,
        *,
        params: Optional[Mapping[str, str]] = None,
        headers: Optional[Mapping[str, str]] = None,
        data: Optional[Union[Mapping[str, str], bytes]] = None,
        json: Any = None,
        files: Optional[RequestFiles] = None,
        timeout: float = -2,
    ) -> Response:
        r = self.make_request(verb, uri, params=params, headers=headers,
                data=data, json=json, files=files)
        return self.send(r, timeout=timeout)

    def close(self) -> None:
        pass

    def _get_effective_timeout_value(self, timeout: float) -> float:
        t1 = self._veto_timeout.get()
        t2 = timeout
        t3 = self.timeout
        for tv in (t1, t2, t3):
            if tv == -2:
                continue
            if 0 > tv != -1:
                raise ValueError('invalid timeout value: %s' % tv)
            return tv
        raise ValueError('a default timeout value could not be determined')

    # Unused for now
    class _TimeoutAsContextManager:
        def __init__(self, var: contextvars.ContextVar[float], timeout: float) -> None:
            self._var = var
            self._timeout = timeout
            self._token_stack: MutableSequence[contextvars.Token[float]] = deque()

        def __enter__(self) -> None:
            tkn = self._var.set(self._timeout)
            self._token_stack.append(tkn)

        def __exit__(self,
            exc_type: Optional[type[BaseException]],
            exc_value: Optional[BaseException],
            exc_traceback: Optional[TracebackType],
        ) -> Optional[bool]:
            tkn = self._token_stack.pop()
            self._var.reset(tkn)
            return None

    # Defer this idea
    def ___timeout_as(self, timeout: float) -> AbstractContextManager[None]:
        return self._TimeoutAsContextManager(self._veto_timeout, timeout)




from __future__ import annotations

from ...core.reddit_http_client_SYNC import RedditHTTPClient
from ...http.util.json_load import json_loads_response
from ...exceptions import (
    ClientException,
    raise_for_reddit_error,
    raise_for_non_json_response,
)


class Procedures:
    def __init__(self, http: RedditHTTPClient):
        self._http = http

    def do_legacy_reddit_web_login(self, username: str, password: str, otp: str = '') -> tuple[str, str]:
        resp = self._http.session.request(
                'POST',
                'https://www.reddit.com/api/login',
                params={'api_type': 'json'},
                headers={
                    'User-Agent': self._http.user_agent,
                    # I don't understand why but the request fails with a `WRONG_PASSWORD` API error if
                    # it includes the `User-Agent` header, but adding some other random header fixes it.
                    'asdf': 'asdf',
                },
                data={'user': username, 'passwd': password, 'otp': otp})
        resp.raise_for_status()

        try:
            json_data = json_loads_response(resp)
        except ValueError as cause:
            try:
                raise_for_non_json_response(resp)
            except Exception as exc:
                raise exc from cause
            raise

        raise_for_reddit_error(json_data)

        d = json_data['json']['data']
        details = d.get('details', '')
        if details == 'TWO_FA_REQUIRED':
            raise ClientException('TWO_FA_REQUIRED')

        return (d['cookie'], d['modhash'])

    def get_sendbird_access_token(self) -> tuple[str, float]:
        resp = self._http.request('GET', 'https://sendbird.reddit.com/api/v1/sendbird/me')
        resp.raise_for_status()
        d = json_loads_response(resp)
        access_token = d['sb_access_token']
        expires_utms = d['valid_until']
        expires_ts = expires_utms / 1000
        return (access_token, expires_ts)




from __future__ import annotations
from typing import TYPE_CHECKING, IO
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.stylesheet import StylesheetInfo

from ...model_loaders.stylesheet import load_stylesheet_info

class LegacyRedditSubredditStyleProcedures:
    def __init__(self, client: Client):
        self._client = client

    async def get_stylesheet(self, sr: str) -> StylesheetInfo:
        root = await self._client.request('GET', '/about/stylesheet', params={'r': sr})
        return load_stylesheet_info(root['data'])

    async def edit_stylesheet(self, sr: str, content: str, *, message: str) -> None:
        data = {'r': sr, 'op': 'save', 'stylesheet_contents': content, 'reason': message}
        await self._client.request('POST', '/api/subreddit_stylesheet', data=data)

    async def upload_stylesheet_image(self, sr: str, file: IO[bytes], name: str) -> None:
        await self._client.request('POST', '/api/upload_sr_img',
                data={'r': sr, 'upload_type': 'img', 'name': name},
                files={'file': file})

    async def delete_stylesheet_image(self, sr: str, name: str) -> None:
        await self._client.request('POST', '/api/delete_sr_img', data={'r': sr, 'img_name': name})

    async def set_icon(self, sr: str, file: IO[bytes]) -> None:
        await self._client.request('POST', '/api/upload_sr_img',
                data={'r': sr, 'upload_type': 'header'},
                files={'file': file})

    async def unset_icon(self, sr: str) -> None:
        await self._client.request('POST', '/api/delete_sr_header', data={'r': sr})

    async def set_mobile_icon(self, sr: str, file: IO[bytes]) -> None:
        await self._client.request('POST', '/api/upload_sr_img',
                data={'r': sr, 'upload_type': 'icon'},
                files={'file': file})

    async def unset_mobile_icon(self, sr: str) -> None:
        await self._client.request('POST', '/api/delete_sr_icon', data={'r': sr})

    async def set_mobile_banner(self, sr: str, file: IO[bytes]) -> None:
        await self._client.request('POST', '/api/upload_sr_img',
                data={'r': sr, 'upload_type': 'banner'},
                files={'file': file})

    async def unset_mobile_banner(self, sr: str) -> None:
        await self._client.request('POST', '/api/delete_sr_banner', data={'r': sr})




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.submission_SYNC import Submission
    from ...models.comment_SYNC import Comment

from functools import cached_property

from ...pagination.paginator_chaining_iterator import ImpartedPaginatorChainingIterator
from ...pagination.paginators.moderation.sync1 import (
    ModQueueListingPaginator,
    ModQueueSubmissionListingPaginator,
    ModQueueCommentListingPaginator,
    ReportsListingPaginator,
    ReportsSubmissionListingPaginator,
    ReportsCommentListingPaginator,
    SpamListingPaginator,
    SpamSubmissionListingPaginator,
    SpamCommentListingPaginator,
    EditedListingPaginator,
    EditedSubmissionListingPaginator,
    EditedCommentListingPaginator,
    UnmoderatedSubmissionListingPaginator,
)

class Pull:
    def __init__(self, client: Client) -> None:
        self._client = client

    class _modqueue:
        def __init__(self, outer: Pull) -> None:
            self._outer = outer
            self._client = outer._client

        def __call__(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingIterator[ModQueueListingPaginator, object]:
            p = ModQueueListingPaginator(self._client, f'/r/{sr}/about/modqueue')
            return ImpartedPaginatorChainingIterator(p, amount)

        def submissions(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingIterator[ModQueueSubmissionListingPaginator, Submission]:
            p = ModQueueSubmissionListingPaginator(self._client, f'/r/{sr}/about/modqueue')
            return ImpartedPaginatorChainingIterator(p, amount)

        def comments(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingIterator[ModQueueCommentListingPaginator, Comment]:
            p = ModQueueCommentListingPaginator(self._client, f'/r/{sr}/about/modqueue')
            return ImpartedPaginatorChainingIterator(p, amount)

    modqueue: cached_property[_modqueue] = cached_property(_modqueue)

    class _reported:
        def __init__(self, outer: Pull) -> None:
            self._outer = outer
            self._client = outer._client

        def __call__(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingIterator[ReportsListingPaginator, object]:
            p = ReportsListingPaginator(self._client, f'/r/{sr}/about/reports')
            return ImpartedPaginatorChainingIterator(p, amount)

        def submissions(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingIterator[ReportsSubmissionListingPaginator, Submission]:
            p = ReportsSubmissionListingPaginator(self._client, f'/r/{sr}/about/reports')
            return ImpartedPaginatorChainingIterator(p, amount)

        def comments(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingIterator[ReportsCommentListingPaginator, Comment]:
            p = ReportsCommentListingPaginator(self._client, f'/r/{sr}/about/reports')
            return ImpartedPaginatorChainingIterator(p, amount)

    reported: cached_property[_reported] = cached_property(_reported)

    class _spam:
        def __init__(self, outer: Pull) -> None:
            self._outer = outer
            self._client = outer._client

        def __call__(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingIterator[SpamListingPaginator, object]:
            p = SpamListingPaginator(self._client, f'/r/{sr}/about/spam')
            return ImpartedPaginatorChainingIterator(p, amount)

        def submissions(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingIterator[SpamSubmissionListingPaginator, Submission]:
            p = SpamSubmissionListingPaginator(self._client, f'/r/{sr}/about/spam')
            return ImpartedPaginatorChainingIterator(p, amount)

        def comments(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingIterator[SpamCommentListingPaginator, Comment]:
            p = SpamCommentListingPaginator(self._client, f'/r/{sr}/about/spam')
            return ImpartedPaginatorChainingIterator(p, amount)

    spam: cached_property[_spam] = cached_property(_spam)

    class _edited:
        def __init__(self, outer: Pull) -> None:
            self._outer = outer
            self._client = outer._client

        def __call__(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingIterator[EditedListingPaginator, object]:
            p = EditedListingPaginator(self._client, f'/r/{sr}/about/edited')
            return ImpartedPaginatorChainingIterator(p, amount)

        def submissions(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingIterator[EditedSubmissionListingPaginator, Submission]:
            p = EditedSubmissionListingPaginator(self._client, f'/r/{sr}/about/edited')
            return ImpartedPaginatorChainingIterator(p, amount)

        def comments(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingIterator[EditedCommentListingPaginator, Comment]:
            p = EditedCommentListingPaginator(self._client, f'/r/{sr}/about/edited')
            return ImpartedPaginatorChainingIterator(p, amount)

    edited: cached_property[_edited] = cached_property(_edited)

    def unmoderated(self, sr: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingIterator[UnmoderatedSubmissionListingPaginator, Submission]:
        p = UnmoderatedSubmissionListingPaginator(self._client, f'/r/{sr}/about/unmoderated')
        return ImpartedPaginatorChainingIterator(p, amount)




from __future__ import annotations
from collections.abc import Iterable

def base_digits(base: int, n: int) -> Iterable[int]:
    if base < 2:
        raise ValueError('`base` must be at least 2')
    if n < 0:
        raise ValueError('`n` must be positive')

    def g(base: int, n: int) -> Iterable[int]:
        while n:
            n, r = divmod(n, base)
            yield r

    return g(base, n)

def to_base(base: int, n: int, alphabet: str = "0123456789abcdefghijklmnopqrstuvwxyz") -> str:
    if base > len(alphabet):
        raise ValueError('alphabet is not large enough')
    if n == 0:
        return alphabet[0]
    sign = '-' if n < 0 else ''
    numbers = base_digits(base, abs(n))
    symbols = (alphabet[i] for i in numbers)
    return sign + ''.join(reversed(list(symbols)))

def to_base36(n: int, alphabet: str = "0123456789abcdefghijklmnopqrstuvwxyz") -> str:
    return to_base(36, n, alphabet=alphabet)




from __future__ import annotations
from typing import TYPE_CHECKING, Mapping, Any
if TYPE_CHECKING:
    from ..client_SYNC import Client

from .modmail_base import (
    BaseConversation,
    BaseMessage,
    BaseModmailModAction,
    BaseUserDossier,
    GBaseConversationAggregate,
    GBaseUserDossierConversationAggregate,
    GBaseOptionalUserDossierConversationAggregate,
)

class Conversation(BaseConversation):
    def __init__(self, d: Mapping[str, Any], client: Client):
        super().__init__(d)
        self.client: Client = client

    def reply(self, body: str, *, hidden: bool = False, internal: bool = False) -> ConversationAggregate:
        return self.client.p.modmail.conversation.reply(self.id, body, hidden=hidden, internal=internal)

    def mark_read(self) -> None:
        self.client.p.modmail.conversation.mark_read(self.id)

    def mark_unread(self) -> None:
        self.client.p.modmail.conversation.mark_unread(self.id)


class Message(BaseMessage):
    def __init__(self, d: Mapping[str, Any], client: Client):
        super().__init__(d)
        self.client: Client = client

class ModmailModAction(BaseModmailModAction):
    pass

class UserDossier(BaseUserDossier):
    pass


class ConversationAggregate(GBaseConversationAggregate[Conversation, Message, ModmailModAction]):
    pass

class UserDossierConversationAggregate(GBaseUserDossierConversationAggregate[Conversation, Message, ModmailModAction, UserDossier]):
    pass

class OptionalUserDossierConversationAggregate(GBaseOptionalUserDossierConversationAggregate[Conversation, Message, ModmailModAction, UserDossier]):
    pass




from __future__ import annotations
from typing import TYPE_CHECKING, MutableMapping, Optional, Sequence, Protocol
if TYPE_CHECKING:
    from importlib.machinery import ModuleSpec

from importlib.util import find_spec
from dataclasses import dataclass

from ...util.imports import load_module_from_spec
from ..websocket_connection_SYNC import WebSocketConnection

class ConnectFunctionProtocol(Protocol):
    def __call__(self, url: str, *,
        subprotocols: Sequence[str] = (), timeout: float = -2
    ) -> WebSocketConnection: ...

@dataclass
class TransportInfo:
    adaptor_module_name: str
    name: str
    version: str
    connect: ConnectFunctionProtocol

def load_spec(name: str, package: Optional[str] = None) -> ModuleSpec:
    module_spec = find_spec(name, package)
    if module_spec is None:
        raise RuntimeError(f'module spec not found: {name} ({package})')
    return module_spec

def load_transport() -> TransportInfo:
    if not transport_info_registry:
        for module_spec in transport_module_spec_list:
            try:
                load_module_from_spec(module_spec)
            except ImportError:
                continue
            break
        else:
            raise ModuleNotFoundError('A websocket transport library needs to be installed.')

    return next(iter(transport_info_registry.values()))

def connect(url: str, *,
        subprotocols: Sequence[str] = (), timeout: float = -2) -> WebSocketConnection:
    connect = load_transport().connect
    return connect(url, subprotocols=subprotocols, timeout=timeout)

def register(
    adaptor_module_name: str,
    name: str,
    version: str,
    connect: ConnectFunctionProtocol,
) -> None:
    info = TransportInfo(
        adaptor_module_name=adaptor_module_name,
        name=name,
        version=version,
        connect=connect,
    )
    transport_info_registry[adaptor_module_name] = info

transport_module_spec_list = [
    load_spec('.carriers.websocket', __package__),
]
transport_info_registry: MutableMapping[str, TransportInfo] = {}




from .token_obtainment_client_ASYNC import TokenObtainmentClient as TokenObtainmentClient  # noqa: F401
from .token_revocation_client_ASYNC import TokenRevocationClient as TokenRevocationClient  # noqa: F401




from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, Sequence
if TYPE_CHECKING:
    from ...client_SYNC import Client

from ...models.user_summary import UserSummary
from ...model_loaders.user_summary import load_user_summary
from ...util.base_conversion import to_base36
from ...iterators.chunking import chunked
from ...iterators.call_chunk_chaining_iterator import CallChunkChainingIterator
from ...iterators.call_chunk import CallChunk
from ... import http

class BulkFetchUserSummary:
    def __init__(self, client: Client):
        self._client = client

    def __call__(self, ids: Iterable[int]) -> CallChunkChainingIterator[UserSummary]:
        def mass_fetch_by_id(ids: Sequence[int]) -> Sequence[UserSummary]:
            id36s = map(to_base36, ids)
            full_id36s = map('t2_'.__add__, id36s)
            ids_str = ','.join(full_id36s)

            try:
                root = self._client.request('GET', '/api/user_data_by_account_ids', params={'ids': ids_str})
            except http.exceptions.StatusCodeException as e:
                if e.status_code == 404:
                    return []
                raise
            return [load_user_summary(v, k) for k, v in root.items()]

        return CallChunkChainingIterator(CallChunk(mass_fetch_by_id, idfs) for idfs in chunked(ids, 100))

    def by_id36(self, id36s: Iterable[str]) -> CallChunkChainingIterator[UserSummary]:
        def mass_fetch_by_id36(id36s: Sequence[str]) -> Sequence[UserSummary]:
            full_id36s = map('t2_'.__add__, id36s)
            ids_str = ','.join(full_id36s)

            try:
                root = self._client.request('GET', '/api/user_data_by_account_ids', params={'ids': ids_str})
            except http.exceptions.StatusCodeException as e:
                if e.status_code == 404:
                    return []
                raise
            return [load_user_summary(v, k) for k, v in root.items()]

        return CallChunkChainingIterator(CallChunk(mass_fetch_by_id36, idfs) for idfs in chunked(id36s, 100))




from __future__ import annotations
from typing import Mapping, Any, Optional, Sequence

from datetime import datetime, timezone

from ..auth.const import AUTHORIZATION_BASE_URL
from .artifact import Artifact
from .report import ModReport, UserReport
from ..model_loaders.report import load_mod_report, load_user_report

class BaseComment(Artifact):
    class Me:
        def __init__(self, d: Mapping[str, Any]):
            self.saved: bool = d['saved']
            self.reply_notifications: bool = d['send_replies']
            self.voted: int = {False: -1, None: 0, True: 1}[d['likes']]

    class Author:
        class AuthorFlair:
            def __init__(self, d: Mapping[str, Any]):
                self.template_uuid: Optional[str] = d['author_flair_template_id']
                author_flair_text: Optional[str] = d['author_flair_text']
                self.text: str = author_flair_text or ''
                self.has_had_flair: bool = author_flair_text is not None
                self.fg_light_or_dark: str = d['author_flair_text_color'] or ''
                self.bg_color: str = d['author_flair_background_color'] or ''
                self.uses_richtext: bool = d['author_flair_type'] == 'richtext'
                author_flair_css_class: Optional[str] = d['author_flair_css_class']
                self.has_had_css_class_when_no_flair_template: bool = author_flair_css_class is not None
                self.css_class: str = author_flair_css_class or ''

        def __init__(self, d: Mapping[str, Any]):
            self.name: str = d['author']
            self.id36: str = d['author_fullname'].split('_', 1)[-1]
            self.id: int = int(self.id36, 36)
            self.has_premium: bool = d['author_premium']
            self.flair: BaseComment.Author.AuthorFlair = self.AuthorFlair(d)

    class Submission:
        def __init__(self, d: Mapping[str, Any]):
            self.id36: str = d['link_id'].split('_', 1)[-1]
            self.id: int = int(self.id36, 36)

    class Subreddit:
        def __init__(self, d: Mapping[str, Any]):
            self.id36: str = d['subreddit_id'].split('_', 1)[-1]
            self.id: int = int(self.id36, 36)
            self.name: str = d['subreddit']
            self.openness: str = d['subreddit_type']

    class Moderator:
        class Approved:
            def __init__(self, d: Mapping[str, Any]):
                self.by: str = d['approved_by']
                self.ut: int = d['approved_at_utc']
                self.at: datetime = datetime.fromtimestamp(self.ut, timezone.utc)

        class Removed:
            def __init__(self, d: Mapping[str, Any]):
                self.by: str = d['banned_by']
                self.ut: int = d['banned_at_utc']
                self.at: datetime = datetime.fromtimestamp(self.ut, timezone.utc)

        class Reports:
            def __init__(self, d: Mapping[str, Any]):
                self.ignoring: bool = d['ignore_reports']
                self.num_reports: int = d['num_reports']
                self.mod_reports: Sequence[ModReport] = [load_mod_report(m) for m in d['mod_reports']]
                self.user_reports: Sequence[UserReport] = [load_user_report(m) for m in d['user_reports']]

        def __init__(self, d: Mapping[str, Any]):
            self.spam: bool = d['spam']

            self.approved: Optional[BaseComment.Moderator.Approved] = None
            if d['approved_by']:
                self.approved = self.Approved(d)

            self.removed: Optional[BaseComment.Moderator.Removed] = None
            if d['banned_by']:
                self.removed = self.Removed(d)

            self.reports: BaseComment.Moderator.Reports = self.Reports(d)

            self.removal_reason_by: Optional[str] = d['mod_reason_by']
            self.removal_reason_title: Optional[str] = d['mod_reason_title']
            self.removal_note: Optional[str] = d['mod_note']

    class _Edited:
        def __init__(self, d: Mapping[str, Any]):
            self.ut: int = d['edited']
            self.at: datetime = datetime.fromtimestamp(self.ut, timezone.utc)

    def __init__(self, d: Mapping[str, Any]):
        super().__init__(d)
        self.id36: str = d['id']
        self.id: int = int(self.id36, 36)
        self.created_ut: int = int(d['created_utc'])
        self.created_at: datetime = datetime.fromtimestamp(self.created_ut, timezone.utc)

        self.body: str = d['body']
        self.body_html: str = d.get('body_html', '')
        #: Works even if score is hidden (`hide_score` JSON field is `True`).
        self.score: int = d['score']
        self.score_hidden: bool = d['score_hidden']

        self.rel_permalink: str = d['permalink']
        self.permalink: str = AUTHORIZATION_BASE_URL + self.rel_permalink

        self.edited: Optional[BaseComment._Edited] = None
        if d['edited']:
            self.edited = self._Edited(d)

        self.is_submitter: bool = d['is_submitter']
        self.stickied: bool = d['stickied']
        self.archived: bool = d['archived']
        self.locked: bool = d['locked']
        self.collapsed: bool = d['collapsed']
        self.distinguished: str = d['distinguished'] or ''

        parent_id: str = d['parent_id']
        self.is_top_level: bool = parent_id.startswith('t3_')
        self.parent_comment_id36: Optional[str] = None
        self.parent_comment_id: Optional[int] = None
        if parent_id.startswith('t1_'):
            self.parent_comment_id36 = parent_id.partition('_')[2]
            self.parent_comment_id = int(self.parent_comment_id36, 36)

        self.me: BaseComment.Me = self.Me(d)

        self.submission: BaseComment.Submission = self.Submission(d)
        self.subreddit: BaseComment.Subreddit = self.Subreddit(d)

        s: str = d['author']
        self.author_name: str = s
        self.author: Optional[BaseComment.Author] = None
        if not s.startswith('['):
            self.author = self.Author(d)

        self.mod: Optional[BaseComment.Moderator] = None
        if 'spam' in d:
            self.mod = self.Moderator(d)


class BaseExtraSubmissionFieldsComment(BaseComment):
    # For:
    # * `GET /comments`
    # * `GET /r/{subreddit}/comments`
    # * `GET /user/{username}/overview` (and variants)

    class Submission2(BaseComment.Submission):
        def __init__(self, d: Mapping[str, Any]):
            super().__init__(d)
            self.title: str = d['link_title']
            self.author_name: str = d['link_author']
            self.rel_permalink: str = d['link_permalink']
            self.permalink: str = AUTHORIZATION_BASE_URL + self.rel_permalink
            self.nsfw: bool = d['over_18']

    class Subreddit2(BaseComment.Subreddit):
        def __init__(self, d: Mapping[str, Any]):
            super().__init__(d)
            self.quarantined: bool = d['quarantine']

    def __init__(self, d: Mapping[str, Any]):
        super().__init__(d)
        self.submission2: BaseExtraSubmissionFieldsComment.Submission2 = self.Submission2(d)
        self.subreddit2: BaseExtraSubmissionFieldsComment.Subreddit2 = self.Subreddit2(d)

class BaseEditPostTextEndpointComment(BaseComment):
    # For: `POST /api/editusertext`
    pass




from __future__ import annotations
from typing import Sequence

from dataclasses import dataclass

@dataclass
class StylesheetImage:
    url: str
    name: str

@dataclass(repr=False, eq=False)
class StylesheetInfo:
    content: str
    images: Sequence[StylesheetImage]




from __future__ import annotations

import sys

from .. import __about__
from ..http.transport.ASYNC import load_transport, transport_registry

def get_user_agent() -> str:
    tt = load_transport()
    return ' '.join([
        f"{__about__.__title__}/{__about__.__version__}",
        f"Python/{sys.version.split(None, 1)[0]}",
        f"{tt.name}/{tt.version}",
    ])

def get_user_agent_from_session(session: object) -> str:
    tokens = [
        f"{__about__.__title__}/{__about__.__version__}",
        f"Python/{sys.version.split(None, 1)[0]}",
    ]
    if session:
        tt = transport_registry.get(session.__module__)
        if tt:
            tokens.append(f"{tt.name}/{tt.version}")
    return ' '.join(tokens)




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.message_ASYNC import MailboxMessage, ComposedMessage, CommentMessage

from ...pagination.paginator_chaining_async_iterator import ImpartedPaginatorChainingAsyncIterator
from ...pagination.paginators.message_async1 import (
    MessageListingAsyncPaginator,
    ComposedMessageListingAsyncPaginator,
    CommentMessageListingAsyncPaginator,
    ComposedMessageThreadListingAsyncPaginator,
)

class Pull:
    def __init__(self, client: Client) -> None:
        self._client = client

    def inbox(self, amount: Optional[int] = None) -> ImpartedPaginatorChainingAsyncIterator[MessageListingAsyncPaginator, MailboxMessage]:
        p = MessageListingAsyncPaginator(self._client, '/message/inbox')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def unread(self, amount: Optional[int] = None) -> ImpartedPaginatorChainingAsyncIterator[MessageListingAsyncPaginator, MailboxMessage]:
        p = MessageListingAsyncPaginator(self._client, '/message/unread')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def messages(self, amount: Optional[int] = None) -> ImpartedPaginatorChainingAsyncIterator[ComposedMessageThreadListingAsyncPaginator, Sequence[ComposedMessage]]:
        p = ComposedMessageThreadListingAsyncPaginator(self._client, '/message/messages')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def sent(self, amount: Optional[int] = None) -> ImpartedPaginatorChainingAsyncIterator[ComposedMessageListingAsyncPaginator, ComposedMessage]:
        p = ComposedMessageListingAsyncPaginator(self._client, '/message/sent')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def comment_replies(self, amount: Optional[int] = None) -> ImpartedPaginatorChainingAsyncIterator[CommentMessageListingAsyncPaginator, CommentMessage]:
        p = CommentMessageListingAsyncPaginator(self._client, '/message/comments')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def post_replies(self, amount: Optional[int] = None) -> ImpartedPaginatorChainingAsyncIterator[CommentMessageListingAsyncPaginator, CommentMessage]:
        p = CommentMessageListingAsyncPaginator(self._client, '/message/selfreply')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def mentions(self, amount: Optional[int] = None) -> ImpartedPaginatorChainingAsyncIterator[CommentMessageListingAsyncPaginator, CommentMessage]:
        p = CommentMessageListingAsyncPaginator(self._client, '/message/mentions')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)




from __future__ import annotations
from typing import TYPE_CHECKING, Mapping, Any
if TYPE_CHECKING:
    from .typedefs import ClientCredentials, AuthorizationGrant
    from ..http.requestor_ASYNC import Requestor

from ..http.request import make_request
from ..http.util.json_load import json_loads_response
from .token import Token
from .utils import apply_basic_auth
from .exceptions import raise_for_token_server_response_error

class TokenObtainmentClient:
    def __init__(self, requestor: Requestor, uri: str,
            client_credentials: ClientCredentials,
            grant: AuthorizationGrant):
        self.requestor: Requestor = requestor
        self.uri: str = uri
        self.client_credentials: tuple[str, str] = client_credentials
        self.grant: Mapping[str, str] = grant

    async def fetch_data(self) -> Mapping[str, Any]:
        r = make_request('POST', self.uri, data=self.grant)
        apply_basic_auth(r, *self.client_credentials)
        resp = await self.requestor.send(r)

        try:
            resp_json = json_loads_response(resp)
        except ValueError:
            resp.raise_for_status()
            raise

        raise_for_token_server_response_error(resp_json)
        resp.raise_for_status()
        return resp_json

    async def fetch_token(self) -> Token:
        return Token.from_dict(await self.fetch_data())




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .....client_ASYNC import Client
    from .....models.submission_ASYNC import Submission
    from .....models.comment_ASYNC import ExtraSubmissionFieldsComment
    from .....models.subreddit_ASYNC import Subreddit
    from ..stream import IStandardStreamEventSubject

from ..stream import Stream


def make_submission_stream(client: Client, sr: str) -> IStandardStreamEventSubject[Submission]:
    it = client.p.subreddit.pull.new(sr)
    paginator = it.get_paginator()
    return Stream(paginator, lambda x: x.id)

def make_comment_stream(client: Client, sr: str) -> IStandardStreamEventSubject[ExtraSubmissionFieldsComment]:
    it = client.p.subreddit.pull_new_comments(sr)
    paginator = it.get_paginator()
    return Stream(paginator, lambda x: x.id)

def make_subreddit_stream(client: Client) -> IStandardStreamEventSubject[Subreddit]:
    it = client.p.subreddit.pulls.new()
    paginator = it.get_paginator()
    return Stream(paginator, lambda x: x.id)




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Mapping, Sequence
if TYPE_CHECKING:
    from ..client_SYNC import Client
    from ..models.message_SYNC import MailboxMessage

from ..models.message_SYNC import ComposedMessage, CommentMessage

def load_composed_message(d: Mapping[str, Any], client: Client) -> ComposedMessage:
    return ComposedMessage(d, client)

def load_comment_message(d: Mapping[str, Any], client: Client) -> CommentMessage:
    return CommentMessage(d, client)

def load_message(d: Mapping[str, Any], client: Client) -> MailboxMessage:
    if d['was_comment']:
        return load_comment_message(d, client)
    return load_composed_message(d, client)

def load_composed_message_thread(d: Mapping[str, Any], client: Client) -> Sequence[ComposedMessage]:
    first = load_composed_message(d, client)
    children = []
    if replies := d['replies']:
        children = [load_composed_message(d['data'], client) for d in replies['data']['children']]
    return [first] + children




from __future__ import annotations
from typing import Any, Mapping, Optional, Sequence

from datetime import datetime, timezone

from ..models.submission_draft import Draft, MarkdownDraft, RichTextDraft


draft_column_names_table: Mapping[str, Sequence[str]] = {
    'id': ('id', 'id'),
    'kind': ('kind', 'kind'),
    'created': ('created', 'created'),
    'modified': ('modified', 'modified'),
    'is_public_link': ('is_public_link', 'isPublicLink'),
    'subreddit': ('subreddit', 'subredditId'),
    'title': ('title', 'title'),
    'body': ('body', 'body'),
    'send_replies': ('send_replies', 'sendReplies'),
    'spoiler': ('spoiler', 'isSpoiler'),
    'nsfw': ('nsfw', 'isNSFW'),
    'original_content': ('original_content', 'isOriginalContent'),
    'flair': ('flair', 'flair'),
}
normal_draft_column_names_map: Mapping[str, str] = {k: v[0] for k, v in draft_column_names_table.items()}
public_draft_column_names_map: Mapping[str, str] = {k: v[1] for k, v in draft_column_names_table.items()}


def _load_draft_impl(d: Mapping[str, Any], h: Mapping[str, str]) -> Draft:
    if d[h['kind']] == 'markdown':
        return load_markdown_draft(d)
    if d[h['kind']] == 'richtext':
        return load_rich_text_draft(d)
    raise ValueError('unknown draft type')

def _construct_draft_impl(d: Mapping[str, Any], h: Mapping[str, str]) -> Draft:
    created_ts: float = d[h['created']] / 1000
    modified_ts: float = d[h['modified']] / 1000

    subreddit_id: Optional[int] = None
    subreddit: Optional[str] = d[h['subreddit']]
    if subreddit is not None:
        _, _, id36 = subreddit.partition('_')
        subreddit_id = int(id36, 36)

    flair: Optional[Draft.FlairInfo] = None
    if df := d[h['flair']]:
        flair = Draft.FlairInfo(
            uuid=df['templateId'],
            type=df['type'],
            text_override=df['text'],
            bg_color=df['backgroundColor'],
            fg_light_or_dark=df['textColor'],
        )

    return Draft(
        d=d,
        uuid=d[h['id']],
        created_at=datetime.fromtimestamp(created_ts, timezone.utc),
        modified_at=datetime.fromtimestamp(modified_ts, timezone.utc),
        public=d[h['is_public_link']],
        subreddit_id=subreddit_id,
        title=d[h['title']],
        reply_notifications=d[h['send_replies']],
        spoiler=d[h['spoiler']],
        nsfw=d[h['nsfw']],
        original_content=d[h['original_content']],
        flair=flair,
    )

def _load_markdown_draft_impl(d: Mapping[str, Any], h: Mapping[str, str]) -> MarkdownDraft:
    u = construct_draft(d)
    return MarkdownDraft(
        d=u.d,
        uuid=u.uuid,
        created_at=u.created_at,
        modified_at=u.modified_at,
        public=u.public,
        subreddit_id=u.subreddit_id,
        title=u.title,
        reply_notifications=u.reply_notifications,
        spoiler=u.spoiler,
        nsfw=u.nsfw,
        original_content=u.original_content,
        flair=u.flair,
        #,
        body=d['body'],
    )

def _load_rich_text_draft_impl(d: Mapping[str, Any], h: Mapping[str, str]) -> RichTextDraft:
    u = construct_draft(d)
    return RichTextDraft(
        d=u.d,
        uuid=u.uuid,
        created_at=u.created_at,
        modified_at=u.modified_at,
        public=u.public,
        subreddit_id=u.subreddit_id,
        title=u.title,
        reply_notifications=u.reply_notifications,
        spoiler=u.spoiler,
        nsfw=u.nsfw,
        original_content=u.original_content,
        flair=u.flair,
    )


def load_draft(d: Mapping[str, Any]) -> Draft:
    return _load_draft_impl(d, normal_draft_column_names_map)

def construct_draft(d: Mapping[str, Any]) -> Draft:
    return _construct_draft_impl(d, normal_draft_column_names_map)

def load_markdown_draft(d: Mapping[str, Any]) -> MarkdownDraft:
    return _load_markdown_draft_impl(d, normal_draft_column_names_map)

def load_rich_text_draft(d: Mapping[str, Any]) -> RichTextDraft:
    return _load_rich_text_draft_impl(d, normal_draft_column_names_map)


def load_public_draft(d: Mapping[str, Any]) -> Draft:
    return _load_draft_impl(d, public_draft_column_names_map)

def construct_public_draft(d: Mapping[str, Any]) -> Draft:
    return _construct_draft_impl(d, public_draft_column_names_map)

def load_public_markdown_draft(d: Mapping[str, Any]) -> MarkdownDraft:
    return _load_markdown_draft_impl(d, public_draft_column_names_map)

def load_public_rich_text_draft(d: Mapping[str, Any]) -> RichTextDraft:
    return _load_rich_text_draft_impl(d, public_draft_column_names_map)



#!/usr/bin/env python3
"""
Step through the OAuth2 Implicit Flow to obtain an access token.

Visit <https://www.reddit.com/prefs/apps> to create an app for your client.
This flow requires an 'installed app' app type.

Your app's redirect URI must *exactly* match `http://localhost:8080`.

Access tokens expire after one hour.
"""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from types import FrameType


import argparse
class Formatter(argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter): pass
parser = argparse.ArgumentParser(description=__doc__, formatter_class=Formatter)
parser._optionals.title = __import__('gettext').gettext('named arguments')
add = parser.add_argument
add('client_id', nargs='?')
add('--client-id', metavar='CLIENT_ID', dest='client_id_opt', help=argparse.SUPPRESS)
add('--scope', default='*', help='an OAuth2 scope string')
add('--redirect-uri', default='http://localhost:8080', help='\N{ZERO WIDTH SPACE}')
add('--duration', choices=['temporary', 'permanent'], default='permanent', help=argparse.SUPPRESS)
add('--no-web-browser', action='store_true', help="don't launch a browser")
add('--web-browser-name', help="the web browser to open")
args = parser.parse_args()


import sys
import os
import uuid
import socket
import urllib.parse
import webbrowser
import signal
from functools import partial

import redditwarp
from redditwarp.http.transport.SYNC import load_transport

def get_client_cred_input(v: Optional[str], prompt: str, env: str) -> str:
    if v is None:
        v = input(prompt)
    if v == '.':
        v = os.environ[env]
        print(v)
    return v

if not sys.flags.interactive:
    @partial(signal.signal, signal.SIGINT)
    def handle_sigint(sig: int, frame: Optional[FrameType]) -> None:
        print('KeyboardInterrupt', file=sys.stderr)
        sys.exit(130)

load_transport()

client_id = get_client_cred_input(
        (args.client_id_opt or args.client_id),
        'Client ID: ', 'redditwarp_client_id')
scope: str = args.scope
redirect_uri: str = args.redirect_uri
duration: str = args.duration
no_web_browser: bool = args.no_web_browser
web_browser_name: str = args.web_browser_name
state = str(uuid.uuid4())

port = urllib.parse.urlsplit(redirect_uri).port
if port is None:
    raise Exception('could not determine the port number from the redirect uri')

browser = webbrowser.get(web_browser_name)

print('\n        -~=~- Reddit OAuth2 Implicit Flow -~=~-\n')
print('Step 1. Build the authorization URL and direct the user to the authorization server.\n')

params = {
    'response_type': 'token',
    'client_id': client_id,
    'redirect_uri': redirect_uri,
    'scope': scope,
    'state': state,
}
url = "%s?%s" % (redditwarp.auth.const.AUTHORIZATION_URL, urllib.parse.urlencode(params))

print(url)
print()

if not no_web_browser:
    browser.open(url)

print('Step 2. Wait for the authorization server response and extract the authorization code.\n')

with socket.socket() as server:
    server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    server.bind(('localhost', port))
    server.listen(1)

    client, addr = server.accept()
    with client:
        data = client.recv(8192)
        client.send(b"HTTP/1.1 200 OK\r\n\r\n" + data)

prompt = 'Enter the address bar URL: \\\n'
while not (redirected_url := input(prompt)):
    pass

urlparts = urllib.parse.urlsplit(redirected_url)
dl = urllib.parse.parse_qs(urlparts.fragment)
response_params = {k: v[0] for k, v in dl.items()}

if state and 'state' not in response_params:
    raise Exception('no state value received')

received_state = response_params['state']
if received_state != state:
    raise Exception(f'sent state ({state}) did not match received ({received_state})')

print()
access_token = response_params['access_token']
print(f'Access token: {access_token}')




from .pull_sync1 import (  # noqa: F401
    OverviewListingPaginator as OverviewListingPaginator,
    CommentsListingPaginator as CommentsListingPaginator,
    SubmittedListingPaginator as SubmittedListingPaginator,
    GildedListingPaginator as GildedListingPaginator,
    UpvotedListingPaginator as UpvotedListingPaginator,
    DownvotedListingPaginator as DownvotedListingPaginator,
    HiddenListingPaginator as HiddenListingPaginator,
    SavedListingPaginator as SavedListingPaginator,
    SavedSubmissionsListingPaginator as SavedSubmissionsListingPaginator,
    SavedCommentsListingPaginator as SavedCommentsListingPaginator,
)
from .pull_user_subreddits_sync1 import UserSubredditListingPaginator as UserSubredditListingPaginator  # noqa: F401
from .explore_sync1 import SearchUsersListingPaginator as SearchUsersListingPaginator  # noqa: F401




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from collections.abc import MutableMapping

from dataclasses import dataclass
from .exceptions import status_successful, raise_now, raise_for_status

@dataclass(eq=False, repr=False)
class Response:
    status: int
    headers: MutableMapping[str, str]
    data: bytes

    @property
    def successful(self) -> bool:
        return self.status_successful()

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} [{self.status}]>'

    def status_successful(self) -> bool:
        return status_successful(self.status)

    def raise_now(self) -> None:
        raise_now(self.status)

    def raise_for_status(self) -> None:
        raise_for_status(self.status)

@dataclass(eq=False, repr=False)
class UResponse(Response):
    underlying_object: object = None

    def __repr__(self) -> str:
        return f'<Response [{self.status}]>'




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .request import Request
    from .response import Response

class Requestor:
    def send(self, request: Request, *, timeout: float = -2) -> Response:
        raise NotImplementedError




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Iterable, Sequence, Tuple
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.flair import FlairTemplate, FlairChoices, UserFlairAssociation

import csv
from io import StringIO

from ...util.base_conversion import to_base36
from ...iterators.chunking import chunked
from ...iterators.call_chunk_chaining_iterator import CallChunkChainingIterator
from ...iterators.call_chunk import CallChunk
from ...pagination.paginator_chaining_iterator import ImpartedPaginatorChainingIterator
from ...pagination.paginators.flair_sync1 import UserFlairAssociationPaginator
from ...model_loaders.flair import (
    load_variant2_flair_template,
    load_variant1_flair_template,
    load_flair_choices,
    load_user_flair_association,
)

class FlairProcedures:
    def __init__(self, client: Client):
        self._client = client

    def assign_user_flair(self, sr_name: str, name: str, text: Optional[str], css_class: Optional[str]) -> None:
        d = {'name': name}
        if text is not None:
            d['text'] = text
        if css_class is not None:
            d['css_class'] = css_class
        self._client.request('POST', f'/r/{sr_name}/api/flair', data=d)

    def assign_post_flair(self, sr_name: str, submission_id: int, text: Optional[str], css_class: Optional[str]) -> None:
        full_id36 = 't3_' + to_base36(submission_id)
        d = {'link': full_id36}
        if text is not None:
            d['text'] = text
        if css_class is not None:
            d['css_class'] = css_class
        self._client.request('POST', f'/r/{sr_name}/api/flair', data=d)

    def revoke_user_flair(self, sr_name: str, name: str) -> None:
        self._client.request('POST', f'/r/{sr_name}/api/flair', data=dict(name=name))

    def revoke_post_flair(self, sr_name: str, submission_id: int) -> None:
        full_id36 = 't3_' + to_base36(submission_id)
        self._client.request('POST', f'/r/{sr_name}/api/flair', data=dict(link=full_id36))

    def bulk_update_user_flairs(self,
        sr_name: str,
        data: Iterable[Tuple[str, str, str]],
    ) -> CallChunkChainingIterator[bool]:
        def mass_update_user_flairs(data: Sequence[Tuple[str, str, str]]) -> Sequence[bool]:
            sio = StringIO()
            csv.writer(sio).writerows(data)
            s = sio.getvalue()
            root = self._client.request('POST', f'/r/{sr_name}/api/flaircsv', data={'flair_csv': s})
            return [i['ok'] for i in root]

        itr = map(
            lambda xs: CallChunk(mass_update_user_flairs, xs),
            chunked(data, 100),
        )
        return CallChunkChainingIterator(itr)

    def _create_or_update_flair_template(self,
        sr_name: str,
        *,
        flair_type: str,
        uuid: Optional[str] = None,
        text: Optional[str] = None,
        allowable_content: Optional[str] = None,
        bg_color: Optional[str] = None,
        css_class: Optional[str] = None,
        fg_light_or_dark: Optional[str] = None,
        text_editable: Optional[bool] = None,
        max_emojis: Optional[int] = None,
        mod_only: Optional[bool] = None,
    ) -> FlairTemplate:
        d = {'flair_type': flair_type}
        for k, v in (
            ('flair_template_id', uuid),
            ('text', text),
            ('allowable_content', allowable_content),
            ('background_color', bg_color),
            ('css_class', css_class),
            ('text_color', fg_light_or_dark),
            ('text_editable', None if text_editable is None else '01'[text_editable]),
            ('max_emojis', str(max_emojis)),
            ('mod_only', None if mod_only is None else '01'[mod_only]),
        ):
            if v is not None:
                d[k] = v
        root = self._client.request('POST', f'/r/{sr_name}/api/flairtemplate_v2', data=d)
        return load_variant2_flair_template(root)

    def create_user_flair_template(self,
        sr_name: str,
        *,
        text: Optional[str] = None,
        allowable_content: Optional[str] = None,
        bg_color: Optional[str] = None,
        css_class: Optional[str] = None,
        fg_light_or_dark: Optional[str] = None,
        text_editable: Optional[bool] = None,
        max_emojis: Optional[int] = None,
        mod_only: Optional[bool] = None,
    ) -> FlairTemplate:
        return self._create_or_update_flair_template(
            sr_name,
            flair_type='USER_FLAIR',
            text=text,
            allowable_content=allowable_content,
            bg_color=bg_color,
            css_class=css_class,
            fg_light_or_dark=fg_light_or_dark,
            text_editable=text_editable,
            max_emojis=max_emojis,
            mod_only=mod_only,
        )

    def create_post_flair_template(self,
        sr_name: str,
        *,
        text: Optional[str] = None,
        allowable_content: Optional[str] = None,
        bg_color: Optional[str] = None,
        css_class: Optional[str] = None,
        fg_light_or_dark: Optional[str] = None,
        text_editable: Optional[bool] = None,
        max_emojis: Optional[int] = None,
        mod_only: Optional[bool] = None,
    ) -> FlairTemplate:
        return self._create_or_update_flair_template(
            sr_name,
            flair_type='POST_FLAIR',
            text=text,
            allowable_content=allowable_content,
            bg_color=bg_color,
            css_class=css_class,
            fg_light_or_dark=fg_light_or_dark,
            text_editable=text_editable,
            max_emojis=max_emojis,
            mod_only=mod_only,
        )

    def update_user_flair_template(self,
        sr_name: str,
        uuid: str,
        *,
        text: Optional[str] = None,
        allowable_content: Optional[str] = None,
        bg_color: Optional[str] = None,
        css_class: Optional[str] = None,
        fg_light_or_dark: Optional[str] = None,
        text_editable: Optional[bool] = None,
        max_emojis: Optional[int] = None,
        mod_only: Optional[bool] = None,
    ) -> FlairTemplate:
        return self._create_or_update_flair_template(
            sr_name,
            flair_type='USER_FLAIR',
            uuid=uuid,
            text=text,
            allowable_content=allowable_content,
            bg_color=bg_color,
            css_class=css_class,
            fg_light_or_dark=fg_light_or_dark,
            text_editable=text_editable,
            max_emojis=max_emojis,
            mod_only=mod_only,
        )

    def update_post_flair_template(self,
        sr_name: str,
        uuid: str,
        *,
        text: Optional[str] = None,
        allowable_content: Optional[str] = None,
        bg_color: Optional[str] = None,
        css_class: Optional[str] = None,
        fg_light_or_dark: Optional[str] = None,
        text_editable: Optional[bool] = None,
        max_emojis: Optional[int] = None,
        mod_only: Optional[bool] = None,
    ) -> FlairTemplate:
        return self._create_or_update_flair_template(
            sr_name,
            flair_type='USER_FLAIR',
            uuid=uuid,
            text=text,
            allowable_content=allowable_content,
            bg_color=bg_color,
            css_class=css_class,
            fg_light_or_dark=fg_light_or_dark,
            text_editable=text_editable,
            max_emojis=max_emojis,
            mod_only=mod_only,
        )

    def assign_user_flair_template(self,
        sr_name: str,
        name: str,
        uuid: Optional[str],
        *,
        text: Optional[str] = None,
        css_class: Optional[str] = None,
        bg_color: Optional[str] = None,
        fg_light_or_dark: Optional[str] = None,
    ) -> None:
        d = {'name': name}
        for k, v in (
            ('flair_template_id', uuid),
            ('text', text),
            ('css_class', css_class),
            ('background_color', bg_color),
            ('text_color', fg_light_or_dark),
        ):
            if v is not None:
                d[k] = v
        self._client.request('POST', f'/r/{sr_name}/api/selectflair', data=d)

    def assign_post_flair_template(self,
        sr_name: str,
        submission_id: int,
        uuid: Optional[str],
        *,
        text: Optional[str] = None,
        css_class: Optional[str] = None,
        bg_color: Optional[str] = None,
        fg_light_or_dark: Optional[str] = None,
    ) -> None:
        full_id36 = 't3_' + to_base36(submission_id)
        d = {'link': full_id36}
        for k, v in (
            ('flair_template_id', uuid),
            ('text', text),
            ('css_class', css_class),
            ('background_color', bg_color),
            ('text_color', fg_light_or_dark),
        ):
            if v is not None:
                d[k] = v
        self._client.request('POST', f'/r/{sr_name}/api/selectflair', data=d)

    def revoke_user_flair_template(self, sr_name: str, name: str) -> None:
        self._client.request('POST', f'/r/{sr_name}/api/selectflair', data=dict(name=name))

    def revoke_post_flair_template(self, sr_name: str, submission_id: int) -> None:
        full_id36 = 't3_' + to_base36(submission_id)
        self._client.request('POST', f'/r/{sr_name}/api/selectflair', data=dict(link=full_id36))

    def delete_user_flair_template(self, sr_name: str, uuid: str) -> None:
        self._client.request('POST', f'/r/{sr_name}/api/deleteflairtemplate',
                data=dict(flair_template_id=uuid))

    def delete_post_flair_template(self, sr_name: str, uuid: str) -> None:
        self.delete_user_flair_template(sr_name, uuid)

    def delete_all_user_flair_templates(self, sr_name: str) -> None:
        self._client.request('POST', f'/r/{sr_name}/api/clearflairtemplates', data={'flair_type': 'USER_FLAIR'})

    def delete_all_post_flair_templates(self, sr_name: str) -> None:
        self._client.request('POST', f'/r/{sr_name}/api/clearflairtemplates', data={'flair_type': 'POST_FLAIR'})

    def configure_subreddit_flair_settings(self,
        sr_name: str,
        *,
        user_flair_enabled: Optional[bool],
        user_flair_position: Optional[str],
        user_flair_self_assign: Optional[bool],
        post_flair_position: Optional[str],
        post_flair_self_assign: Optional[bool],
    ) -> None:
        def g() -> Iterable[tuple[str, str]]:
            if user_flair_enabled is not None:
                yield ('flair_enabled', '01'[user_flair_enabled])
            if user_flair_position is not None:
                yield ('flair_position', user_flair_position)
            if user_flair_self_assign is not None:
                yield ('flair_self_assign_enabled', '01'[user_flair_self_assign])
            if post_flair_position is not None:
                yield ('link_flair_position', post_flair_position)
            if post_flair_self_assign is not None:
                yield ('link_flair_self_assign_enabled', '01'[post_flair_self_assign])

        self._client.request('POST', f'/r/{sr_name}/api/flairconfig', data=dict(g()))

    def reorder_user_flair_templates(self, sr_name: str, order: Sequence[str]) -> None:
        params = {'subreddit': sr_name, 'flair_type': 'USER_FLAIR'}
        self._client.request('PATCH', '/api/flair_template_order', params=params, json=order)

    def reorder_post_flair_templates(self, sr_name: str, order: Sequence[str]) -> None:
        params = {'subreddit': sr_name, 'flair_type': 'POST_FLAIR'}
        self._client.request('PATCH', '/api/flair_template_order', params=params, json=order)

    def get_user_flair_templates(self, sr_name: str) -> Sequence[FlairTemplate]:
        root = self._client.request('GET', f'/r/{sr_name}/api/user_flair_v2')
        return [load_variant1_flair_template(i) for i in root]

    def get_post_flair_templates(self, sr_name: str) -> Sequence[FlairTemplate]:
        root = self._client.request('GET', f'/r/{sr_name}/api/link_flair_v2')
        return [load_variant1_flair_template(i) for i in root]

    def get_user_flair_choices(self, sr_name: str, name: Optional[str]) -> Optional[FlairChoices]:
        d = {}
        if name is not None:
            d['name'] = name
        root = self._client.request('POST', f'/r/{sr_name}/api/flairselector', data=d)
        if root == '{}':
            return None
        return load_flair_choices(root)

    def get_post_flair_choices(self, sr_name: str, submission_id: Optional[int]) -> Optional[FlairChoices]:
        d = {'is_newlink': '1'}
        if submission_id is not None:
            full_id36 = to_base36(submission_id)
            d['link'] = full_id36
        root = self._client.request('POST', f'/r/{sr_name}/api/flairselector', data=d)
        if root == '{}':
            return None
        return load_flair_choices(root)

    def get_user_flair_association(self, sr_name: str, name: str) -> Optional[UserFlairAssociation]:
        params = {'name': name, 'limit': '1'}
        root = self._client.request('GET', f'/r/{sr_name}/api/flairlist', params=params)
        users_data = root['users']
        if not users_data:
            return None
        user_data = users_data[0]
        if user_data['user'].lower() != name.lower():
            return None
        return load_user_flair_association(user_data)

    def get_user_flair_associations(self,
        sr_name: str,
        amount: Optional[int] = None,
    ) -> ImpartedPaginatorChainingIterator[UserFlairAssociationPaginator, UserFlairAssociation]:
        p = UserFlairAssociationPaginator(self._client, f'/r/{sr_name}/api/flairlist')
        return ImpartedPaginatorChainingIterator(p, amount)

    def show_my_flair(self, sr_name: str) -> None:
        self._client.request('GET', f'/r/{sr_name}/api/setflairenabled', params={'flair_enabled': '1'})

    def hide_my_flair(self, sr_name: str) -> None:
        self._client.request('GET', f'/r/{sr_name}/api/setflairenabled', params={'flair_enabled': '0'})




from __future__ import annotations
from typing import Any, Mapping, Iterator

class PushshiftDocument(Mapping[str, Any]):
    def __init__(self, d: Mapping[str, Any]) -> None:
        self.d: Mapping[str, Any] = d
        # self.id: int = int(d['id'], 36)

    def __contains__(self, item: object) -> bool:
        return item in self.d
    def __iter__(self) -> Iterator[str]:
        return iter(self.d)
    def __len__(self) -> int:
        return len(self.d)
    def __getitem__(self, key: str) -> Any:
        return self.d[key]




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence, Iterable
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.moderation_note import ModerationUserNote

from ...pagination.paginator_chaining_async_iterator import ImpartedPaginatorChainingAsyncIterator
from ...pagination.paginators.moderation.note_pulls_async1 import ModerationNoteAsyncPaginator
from ...iterators.chunking import chunked
from ...iterators.call_chunk_chaining_async_iterator import CallChunkChainingAsyncIterator
from ...iterators.async_call_chunk import AsyncCallChunk
from ...model_loaders.moderation_note import load_moderation_user_note
from ... import http

class Note:
    def __init__(self, client: Client) -> None:
        self._client = client

    async def create_user_note(self, subreddit: str, user: str, note: str, *,
            label: str = '',
            anchor_submission_id: Optional[int] = None,
            anchor_comment_id: Optional[int] = None) -> ModerationUserNote:
        mep = (anchor_submission_id, anchor_comment_id)
        if len(mep) - mep.count(None) > 1:
            raise TypeError('mutually exclusive parameters: `anchor_submission_id`, `anchor_comment_id)`.')

        params = {'subreddit': subreddit, 'user': user, 'label': label}
        root = await self._client.request('POST', '/api/mod/notes', params=params)
        return load_moderation_user_note(root['created'])

    def pulls(self,
        subreddit: str,
        user: str,
        amount: Optional[int] = None,
        *,
        type: Optional[str] = None,
    ) -> ImpartedPaginatorChainingAsyncIterator[ModerationNoteAsyncPaginator, ModerationUserNote]:
        p = ModerationNoteAsyncPaginator(
            client=self._client,
            uri='/api/mod/notes',
            subreddit=subreddit,
            user=user,
            type=type,
        )
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    async def delete(self, note_uuid: str, subreddit: str, user: str) -> None:
        params = {'note_id': 'ModNote_' + note_uuid, 'subreddit': subreddit, 'user': user}
        await self._client.request('DELETE', '/api/mod/notes', params=params)

    async def get_latest_user_note(self, subreddit: str, user: str) -> Optional[ModerationUserNote]:
        params = {'subreddits': subreddit, 'users': user}
        try:
            root = await self._client.request('GET', '/api/mod/notes/recent', params=params)
        except http.exceptions.StatusCodeException as e:
            if e.status_code == 400:
                return None
            raise
        return load_moderation_user_note(root['mod_notes'][0])

    def bulk_get_latest_user_note(self, pairs: Iterable[tuple[str, str]]) -> CallChunkChainingAsyncIterator[Optional[ModerationUserNote]]:
        async def mass_get_latest_user_note(pairs: Sequence[tuple[str, str]]) -> Sequence[Optional[ModerationUserNote]]:
            subreddits, users = tuple(zip(*pairs))
            subreddits_str = ','.join(subreddits)
            users_str = ','.join(users)
            params = {'subreddits': subreddits_str, 'users': users_str}
            try:
                root = await self._client.request('GET', '/api/mod/notes/recent', params=params)
            except http.exceptions.StatusCodeException as e:
                if e.status_code == 400:
                    return [None]
                raise
            return [
                (None if note_obj is None else load_moderation_user_note(note_obj))
                for note_obj in root['mod_notes']
            ]

        return CallChunkChainingAsyncIterator(AsyncCallChunk(mass_get_latest_user_note, chunk) for chunk in chunked(pairs, 500))




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Mapping
if TYPE_CHECKING:
    from ..client_ASYNC import Client

from ..models.comment_ASYNC import (
    Comment,
    ExtraSubmissionFieldsComment,
    EditPostTextEndpointComment,
)

def load_comment(d: Mapping[str, Any], client: Client) -> Comment:
    return Comment(d, client)

def load_extra_submission_fields_comment(d: Mapping[str, Any], client: Client) -> ExtraSubmissionFieldsComment:
    return ExtraSubmissionFieldsComment(d, client)

def load_edit_post_text_endpoint_comment(d: Mapping[str, Any], client: Client) -> EditPostTextEndpointComment:
    return EditPostTextEndpointComment(d, client)




from __future__ import annotations
from typing import Any, Mapping

from ..models.moderator_list import ModeratorListItem

def load_moderator_list_item(d: Mapping[str, Any]) -> ModeratorListItem:
    return ModeratorListItem(d)



#!/usr/bin/env python3
"""
Revoke an access or refresh token.

If neither -a nor -r is specified then the token server will automatically
determine the token type, as per RFC 7009 (Section 2.1).
"""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from types import FrameType


import argparse
class Formatter(argparse.RawDescriptionHelpFormatter): pass
parser = argparse.ArgumentParser(description=__doc__, formatter_class=Formatter)
parser._optionals.title = __import__('gettext').gettext('named arguments')
group = parser.add_mutually_exclusive_group()
group.add_argument('-a', action='store_true', help="hint to the server that the token is an access token")
group.add_argument('-r', action='store_true', help="hint to the server that the token is a refresh token")
parser.add_argument('client_id', nargs='?')
parser.add_argument('client_secret', nargs='?')
parser.add_argument('token', nargs='?')
parser.add_argument('--client-id', metavar='CLIENT_ID', dest='client_id_opt', help=argparse.SUPPRESS)
parser.add_argument('--client-secret', metavar='CLIENT_SECRET', dest='client_secret_opt', help=argparse.SUPPRESS)
args = parser.parse_args()


import sys
import os
import signal
from functools import partial

import redditwarp
from redditwarp.http.transport.SYNC import load_transport, new_session
from redditwarp.auth.SYNC import TokenRevocationClient
from redditwarp.http.misc.apply_params_and_headers_SYNC import ApplyDefaultParamsAndHeaders
from redditwarp.util.user_agent_SYNC import get_user_agent_from_session

def get_client_cred_input(v: Optional[str], prompt: str, env: str) -> str:
    if v is None:
        v = input(prompt)
    if v == '.':
        v = os.environ[env]
        print(v)
    return v

if not sys.flags.interactive:
    @partial(signal.signal, signal.SIGINT)
    def handle_sigint(sig: int, frame: Optional[FrameType]) -> None:
        print('KeyboardInterrupt', file=sys.stderr)
        sys.exit(130)

load_transport()

client_id = get_client_cred_input(
        (args.client_id_opt or args.client_id),
        'Client ID: ', 'redditwarp_client_id')
client_secret = get_client_cred_input(
        (args.client_secret_opt or args.client_secret),
        'Client secret: ', 'redditwarp_client_secret')
token: str = args.token or input('Token: ')
access_token_needs_revoking: bool = args.a
refresh_token_needs_revoking: bool = args.r

session = new_session()
ua = get_user_agent_from_session(session) + " redditwarp.cli.revoke_token"
requestor = ApplyDefaultParamsAndHeaders(session, headers={'User-Agent': ua})
revoke_token_client = TokenRevocationClient(
    requestor,
    redditwarp.auth.const.TOKEN_REVOCATION_URL,
    (client_id, client_secret),
)

if access_token_needs_revoking == refresh_token_needs_revoking:
    revoke_token_client.revoke_token(token)
elif access_token_needs_revoking:
    revoke_token_client.revoke_access_token(token)
else:
    revoke_token_client.revoke_refresh_token(token)




from __future__ import annotations
from typing import Mapping, Any, Sequence, Iterator, Union, overload

class Contributor:
    def __init__(self, d: Mapping[str, Any]):
        self.d: Mapping[str, Any] = d
        self.name: str = d['name']
        full_id36: str = d['id']
        _, _, id36 = full_id36.partition('_')
        self.id36: str = id36
        self.id: int = int(id36, 36)
        self.permissions: Sequence[str] = d['permissions']

class ContributorList(Sequence[Contributor]):
    def __init__(self,
            contributors: Sequence[Contributor],
            invitations: Sequence[Contributor]):
        self.contributors: Sequence[Contributor] = contributors
        self.invitations: Sequence[Contributor] = invitations

    def __len__(self) -> int:
        return len(self.contributors)

    def __contains__(self, item: object) -> bool:
        return item in self.contributors

    def __iter__(self) -> Iterator[Contributor]:
        return iter(self.contributors)

    @overload
    def __getitem__(self, index: int) -> Contributor: ...
    @overload
    def __getitem__(self, index: slice) -> Sequence[Contributor]: ...
    def __getitem__(self, index: Union[int, slice]) -> Union[Contributor, Sequence[Contributor]]:
        return self.contributors[index]




from __future__ import annotations
from typing import Sequence

from dataclasses import dataclass

@dataclass(repr=False, eq=False)
class Tab:
    label: str

@dataclass(repr=False, eq=False)
class LinkTab(Tab):
    link: str


@dataclass(repr=False, eq=False)
class SubmenuItem:
    label: str
    link: str

@dataclass(repr=False, eq=False)
class SubmenuTab(Tab):
    items: Sequence[SubmenuItem]




from __future__ import annotations
from typing import Any, Mapping, Optional, cast

from functools import cached_property
from datetime import datetime, timezone

class SubredditUserItem:
    @cached_property
    def added_at(self) -> datetime:
        return datetime.fromtimestamp(self.added_ut, timezone.utc)

    def __init__(self, d: Mapping[str, Any]) -> None:
        self.d: Mapping[str, Any] = d
        full_id36: str = d['id']
        _, _, id36 = full_id36.partition('_')
        self.id36: str = id36
        self.id: int = int(id36, 36)
        self.name: str = d['username']
        self.account_icon: str = d['accountIcon']
        self.icon_size: tuple[int, int] = cast("tuple[int, int]", d['iconSize'])
        self.added_ut: int = 0

class ModeratorUserItem(SubredditUserItem):
    def __init__(self, d: Mapping[str, Any]) -> None:
        super().__init__(d)
        self.added_ut: int = d['moddedAtUTC']
        self.flair_text: str = d['authorFlairText']
        self.permissions: set[str] = {k for k, v in d['modPermissions'].items() if v}
        self.post_karma: int = d['postKarma']

class ContributorUserItem(SubredditUserItem):
    def __init__(self, d: Mapping[str, Any]) -> None:
        super().__init__(d)
        self.added_ut: int = d['approvedAtUTC']

class BannedUserItem(SubredditUserItem):
    def __init__(self, d: Mapping[str, Any]) -> None:
        super().__init__(d)
        self.added_ut: int = d['bannedAtUTC']
        self.agent_name: str = d['bannedBy']
        self.reason: str = d['reason'] or ''
        self.note: str = d['modNote'] or ''
        self.ban_message: str = d['banMessage']
        self.days_remaining: Optional[int] = d['duration']

class MutedUserItem(SubredditUserItem):
    def __init__(self, d: Mapping[str, Any]) -> None:
        super().__init__(d)
        self.added_ut: int = d['mutedAtUTC']
        self.agent_name: str = d['mutedBy']
        self.reason: str = d['reason']




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.submission_ASYNC import Submission
    from ...models.comment_ASYNC import Comment

from functools import cached_property

from ...pagination.paginator_chaining_async_iterator import ImpartedPaginatorChainingAsyncIterator
from ...pagination.paginators.moderation.async1 import (
    ModQueueListingAsyncPaginator,
    ModQueueSubmissionListingAsyncPaginator,
    ModQueueCommentListingAsyncPaginator,
    ReportsListingAsyncPaginator,
    ReportsSubmissionListingAsyncPaginator,
    ReportsCommentListingAsyncPaginator,
    SpamListingAsyncPaginator,
    SpamSubmissionListingAsyncPaginator,
    SpamCommentListingAsyncPaginator,
    EditedListingAsyncPaginator,
    EditedSubmissionListingAsyncPaginator,
    EditedCommentListingAsyncPaginator,
    UnmoderatedSubmissionListingAsyncPaginator,
)

class Pull:
    def __init__(self, client: Client) -> None:
        self._client = client

    class _modqueue:
        def __init__(self, outer: Pull) -> None:
            self._outer = outer
            self._client = outer._client

        def __call__(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingAsyncIterator[ModQueueListingAsyncPaginator, object]:
            p = ModQueueListingAsyncPaginator(self._client, f'/r/{sr}/about/modqueue')
            return ImpartedPaginatorChainingAsyncIterator(p, amount)

        def submissions(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingAsyncIterator[ModQueueSubmissionListingAsyncPaginator, Submission]:
            p = ModQueueSubmissionListingAsyncPaginator(self._client, f'/r/{sr}/about/modqueue')
            return ImpartedPaginatorChainingAsyncIterator(p, amount)

        def comments(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingAsyncIterator[ModQueueCommentListingAsyncPaginator, Comment]:
            p = ModQueueCommentListingAsyncPaginator(self._client, f'/r/{sr}/about/modqueue')
            return ImpartedPaginatorChainingAsyncIterator(p, amount)

    modqueue: cached_property[_modqueue] = cached_property(_modqueue)

    class _reported:
        def __init__(self, outer: Pull) -> None:
            self._outer = outer
            self._client = outer._client

        def __call__(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingAsyncIterator[ReportsListingAsyncPaginator, object]:
            p = ReportsListingAsyncPaginator(self._client, f'/r/{sr}/about/reports')
            return ImpartedPaginatorChainingAsyncIterator(p, amount)

        def submissions(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingAsyncIterator[ReportsSubmissionListingAsyncPaginator, Submission]:
            p = ReportsSubmissionListingAsyncPaginator(self._client, f'/r/{sr}/about/reports')
            return ImpartedPaginatorChainingAsyncIterator(p, amount)

        def comments(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingAsyncIterator[ReportsCommentListingAsyncPaginator, Comment]:
            p = ReportsCommentListingAsyncPaginator(self._client, f'/r/{sr}/about/reports')
            return ImpartedPaginatorChainingAsyncIterator(p, amount)

    reported: cached_property[_reported] = cached_property(_reported)

    class _spam:
        def __init__(self, outer: Pull) -> None:
            self._outer = outer
            self._client = outer._client

        def __call__(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingAsyncIterator[SpamListingAsyncPaginator, object]:
            p = SpamListingAsyncPaginator(self._client, f'/r/{sr}/about/spam')
            return ImpartedPaginatorChainingAsyncIterator(p, amount)

        def submissions(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingAsyncIterator[SpamSubmissionListingAsyncPaginator, Submission]:
            p = SpamSubmissionListingAsyncPaginator(self._client, f'/r/{sr}/about/spam')
            return ImpartedPaginatorChainingAsyncIterator(p, amount)

        def comments(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingAsyncIterator[SpamCommentListingAsyncPaginator, Comment]:
            p = SpamCommentListingAsyncPaginator(self._client, f'/r/{sr}/about/spam')
            return ImpartedPaginatorChainingAsyncIterator(p, amount)

    spam: cached_property[_spam] = cached_property(_spam)

    class _edited:
        def __init__(self, outer: Pull) -> None:
            self._outer = outer
            self._client = outer._client

        def __call__(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingAsyncIterator[EditedListingAsyncPaginator, object]:
            p = EditedListingAsyncPaginator(self._client, f'/r/{sr}/about/edited')
            return ImpartedPaginatorChainingAsyncIterator(p, amount)

        def submissions(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingAsyncIterator[EditedSubmissionListingAsyncPaginator, Submission]:
            p = EditedSubmissionListingAsyncPaginator(self._client, f'/r/{sr}/about/edited')
            return ImpartedPaginatorChainingAsyncIterator(p, amount)

        def comments(self, sr: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingAsyncIterator[EditedCommentListingAsyncPaginator, Comment]:
            p = EditedCommentListingAsyncPaginator(self._client, f'/r/{sr}/about/edited')
            return ImpartedPaginatorChainingAsyncIterator(p, amount)

    edited: cached_property[_edited] = cached_property(_edited)

    def unmoderated(self, sr: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingAsyncIterator[UnmoderatedSubmissionListingAsyncPaginator, Submission]:
        p = UnmoderatedSubmissionListingAsyncPaginator(self._client, f'/r/{sr}/about/unmoderated')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)




from __future__ import annotations
from typing import Mapping, Any, Optional

from datetime import datetime
from dataclasses import dataclass

from .artifact import IArtifact

@dataclass(repr=False, eq=False)
class Draft(IArtifact):
    @dataclass(repr=False, eq=False)
    class FlairInfo:
        uuid: str = ''
        type: str = ''
        text_override: str = ''
        bg_color: str = ''
        fg_light_or_dark: str = ''

    d: Mapping[str, Any]
    uuid: str
    created_at: datetime
    modified_at: datetime
    public: bool
    subreddit_id: Optional[int]
    title: str
    reply_notifications: bool
    spoiler: bool
    nsfw: bool
    original_content: bool
    flair: Optional[FlairInfo]

@dataclass(repr=False, eq=False)
class MarkdownDraft(Draft):
    body: str

@dataclass(repr=False, eq=False)
class RichTextDraft(Draft):
    pass




from __future__ import annotations

from typing import TYPE_CHECKING, Optional, ContextManager
if TYPE_CHECKING:
    from types import TracebackType

class except_without_context(ContextManager[bool]):
    """
    Example 1::

        def f() -> None:
            try:
                raise LookupError
            except LookupError as e:
                raise KeyError from e

        try:
            raise PermissionError
        except PermissionError as e:
            f()

    Result::

        PermissionError
        During handling of the above exception, another exception occurred:
        LookupError
        The above exception was the direct cause of the following exception:
        KeyError

    We don't want PermissionError to show.


    Example 2::

        try:
            raise PermissionError
        except PermissionError as e:
            try:
                f()
            except Exception as e:
                raise e from None

    Result::

        KeyError

    LookupError is missing.


    Example 3::

        with except_without_context(PermissionError) as ewc:
            raise PermissionError
        if ewc:
            f()

    Result::

        LookupError
        The above exception was the direct cause of the following exception:
        KeyError

    Just right.
    """

    def __init__(self, *exceptions: type[BaseException]):
        self.exceptions: tuple[type[BaseException], ...] = exceptions
        self.yes: bool = False

    def __bool__(self) -> bool:
        return self.yes

    def __exit__(self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        exc_traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        self.yes = exc_type is not None and issubclass(exc_type, self.exceptions)
        return True




from __future__ import annotations
from typing import TYPE_CHECKING, TypeVar, Any, Mapping, Optional, Callable, Iterable
if TYPE_CHECKING:
    from ....client_SYNC import Client

from ...paginator import CursorPaginator, Bidirectional, MoreAvailablePaginator, Resettable

T = TypeVar('T')

class ListingPaginator(Resettable, MoreAvailablePaginator[T], Bidirectional, CursorPaginator[T]):
    def __init__(self,
        client: Client,
        uri: str,
        *,
        limit: Optional[int] = 100,
        params: Optional[Mapping[str, str]] = None,
        cursor_extractor: Callable[[Any], str] = lambda x: x['data']['name'],
    ):
        super().__init__(limit=limit)
        self.client: Client = client
        self.uri: str = uri
        self.params: Mapping[str, str] = {} if params is None else params
        self.cursor_extractor: Callable[[Any], str] = cursor_extractor

        self.direction: bool = True

        self.show_all: bool = False

        self.__reset()

    def __reset(self) -> None:
        self.after: str = ''
        self.before: str = ''
        self.has_after: bool = True
        self.has_before: bool = True
        self.after_count: int = 0
        self.before_count: int = 0

    def reset(self) -> None:
        self.__reset()

    def get_cursor(self) -> str:
        return self.after if self.direction else self.before

    def set_cursor(self, value: str) -> None:
        if self.direction:
            self.after = value
        else:
            self.before = value

    def more_available(self) -> bool:
        return self.has_after if self.direction else self.has_before

    def set_more_available_flag(self, value: bool) -> None:
        if self.direction:
            self.has_after = value
        else:
            self.has_before = value

    def _generate_params(self) -> Iterable[tuple[str, str]]:
        yield from self.params.items()

        if self.limit is not None:
            yield ('limit', str(self.limit))
        if self.show_all:
            yield ('show', 'all')

        if self.direction:
            if self.after_count:
                yield ('count', str(self.after_count))
            if self.after:
                yield ('after', self.after)
        else:
            if self.before_count:
                yield ('count', str(self.before_count))
            if self.before:
                yield ('before', self.before)

    def _fetch_data(self) -> Any:
        params = dict(self._generate_params())
        root = self.client.request('GET', self.uri, params=params)
        data = root['data']
        children = data['children']

        dist: int = x if (x := data['dist']) else len(children)
        self.after_count += (1 if self.direction else -1) * dist
        self.before_count = self.after_count - dist + 1

        suggested_forward_cursor = data['after'] or ''
        suggested_backward_cursor = data['before'] or ''
        if children:
            self.after = suggested_forward_cursor if suggested_forward_cursor else self.cursor_extractor(children[-1])
            self.before = suggested_backward_cursor if suggested_backward_cursor else self.cursor_extractor(children[0])
        self.has_after = bool(suggested_forward_cursor)
        self.has_before = bool(suggested_backward_cursor)

        return data




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Iterable
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.submission_draft import Draft
    from ...models.submission_draft_SYNC import DraftList

from ...model_loaders.submission_draft import load_public_draft
from ...model_loaders.submission_draft_SYNC import load_draft_list
from ...util.base_conversion import to_base36

class DraftProcedures:
    def __init__(self, client: Client):
        self._client = client

    def create_markdown_draft(self,
        *,
        public: Optional[bool] = None,
        subreddit_id: Optional[int] = None,
        title: Optional[str] = None,
        body: Optional[str] = None,
        reply_notifications: Optional[bool] = None,
        spoiler: Optional[bool] = None,
        nsfw: Optional[bool] = None,
        original_content: Optional[bool] = None,
        flair_text_override: Optional[str] = None,
    ) -> str:
        def g() -> Iterable[tuple[str, str]]:
            yield ('kind', 'markdown')
            if public is not None: yield ('is_public_link', '01'[public])
            if subreddit_id is not None: yield ('subreddit', 't5_' + to_base36(subreddit_id))
            if title is not None: yield ('title', title)
            if body is not None: yield ('body', body)
            if reply_notifications is not None: yield ('send_replies', '01'[reply_notifications])
            if spoiler is not None: yield ('spoiler', '01'[spoiler])
            if nsfw is not None: yield ('nsfw', '01'[nsfw])
            if original_content is not None: yield ('original_content', '01'[original_content])
            if flair_text_override is not None: yield ('flair_text', flair_text_override)

        root = self._client.request('POST', '/api/v1/draft', data=dict(g()))
        return root['json']['data']['id']

    def retrieve_my_drafts(self) -> DraftList:
        root = self._client.request('GET', '/api/v1/drafts')
        return load_draft_list(root, self._client)

    def read_public_draft(self, user: str, draft_uuid: str) -> Draft:
        uri = f"https://gateway.reddit.com/desktopapi/v1/draftpreviewpage/{user}/{draft_uuid}"
        root = self._client.request('GET', uri)
        draft_data = root['drafts'][draft_uuid]
        return load_public_draft(draft_data)

    def update_markdown_draft(self,
        uuid: str,
        *,
        public: Optional[bool] = None,
        subreddit_id: Optional[int] = None,
        title: Optional[str] = None,
        body: Optional[str] = None,
        reply_notifications: Optional[bool] = None,
        spoiler: Optional[bool] = None,
        nsfw: Optional[bool] = None,
        original_content: Optional[bool] = None,
        flair_text_override: Optional[str] = None,
    ) -> str:
        def g() -> Iterable[tuple[str, str]]:
            yield ('id', 'uuid')
            yield ('kind', 'markdown')
            if public is not None: yield ('is_public_link', '01'[public])
            if subreddit_id is not None: yield ('subreddit', 't5_' + to_base36(subreddit_id))
            if title is not None: yield ('title', title)
            if body is not None: yield ('body', body)
            if reply_notifications is not None: yield ('send_replies', '01'[reply_notifications])
            if spoiler is not None: yield ('spoiler', '01'[spoiler])
            if nsfw is not None: yield ('nsfw', '01'[nsfw])
            if original_content is not None: yield ('original_content', '01'[original_content])
            if flair_text_override is not None: yield ('flair_text', flair_text_override)

        root = self._client.request('PUT', '/api/v1/draft', data=dict(g()))
        return root['json']['data']['id']

    def delete(self, uuid: str) -> None:
        self._client.request('DELETE', '/api/v1/draft', params={'draft_id': uuid})
